/**
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 * Authors:
 * 2009-
 *	Oscar Koeroo <okoeroo@nikhef.nl>
 *	Mischa Sall\'e <msalle@nikhef.nl>
 *	NIKHEF Amsterdam, the Netherlands
 * 2006-2009
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 */


/*!
 *  \file glexec_conf.c
 *
 */

#include "glexec_conf.h"

#include "glexec.h"
#include "glexec_util.h"
#include "glexec_log.h"
#include "fileutil/fileutil.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <grp.h>
#include <libgen.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/*************************************************************************/
/* DEFINES/TYPEDEFS                                                      */
/*************************************************************************/

#define CONF_ELEM_MAX	199	/* Maximum number of entries in config list */

/*!
 *  Error codes that the configuration can generate.
 */
typedef enum glexec_conf_error_s {
  GLEXEC_CONF_SUCCESS                 = 100,  /*!<  All is well. */
  GLEXEC_CONF_COULD_NOT_OPEN_FILE     = 101,  /*!<  The requested file could not be opened. */
  GLEXEC_CONF_PARSING_FAILED          = 102,  /*!<  An error occurred during parsing. */
  GLEXEC_CONF_LABEL_NOT_CLOSED        = 103,  /*!<  The closing bracket ']' was not found. */
  GLEXEC_CONF_OUT_OF_MEMORY           = 104,  /*!<  There seems to be a shortage on memory. */
  GLEXEC_CONF_ERROR_IN_INI_FILE       = 105,  /*!<  The format of the .ini is broken. */
  GLEXEC_CONF_DOUBLE_KEY              = 106,  /*!<  Key for current section has been defined already. */
  GLEXEC_CONF_NEEDS_UNPRIVILIGED_USER = 107,  /*!<  Check for unprivileged user failed. */
  GLEXEC_CONF_UNINITIALIZED           = 108,  /*!<  The internal representation of the config file is not initialised yet. */
  GLEXEC_CONF_STAT_FAILED             = 109,  /*!<  Stat has failed. */
} glexec_conf_error_t;

/*!
 *  \brief File header and line number info.
 *  Hold the file header of the file that is being parsed and the current line number.
 */
typedef struct glexec_conf_parse_file_s {
  unsigned int lineno;         /*!<  Current line number of the parsed file.*/
  char *buffer;		       /*!<  buffer. */
  int index;		       /*!<  index within the buffer. */
  const char *name;            /*!<  Name (full path) of the configuration file.*/
} glexec_conf_parse_file_t;


/*!
 *  \brief Single linked list of key value pairs.
 *
 *  Group a key and its value together and remember on which line in
 *  the configuration file they were found. All glexec_conf_pair_s
 *  form a single linked list.
 */
typedef struct glexec_conf_pair_s {
  unsigned int lineno;               /*!<  Line number of the key value pair.*/
  char *key;                         /*!<  Key.*/
  char *value;                       /*!<  Value associated with the key.*/
  struct glexec_conf_pair_s *next;   /*!<  Next key value pair.*/
} glexec_conf_pair_t;


/*!
 *  \brief Internal representation of a config file.
 *
 *  The internal representation of the configuration file is a hash table for
 *  quick looking up key value pairs. The key determines the hash. The size
 *  of the table can be determined at run time. \see glexec_conf_init().
 */
typedef struct glexec_config_s {
  int table_size;               /*!<  Number of elements of the table.*/
  glexec_conf_pair_t **list;    /*!<  All the elements of the table.*/
} glexec_config_t;


/*************************************************************************/
/* GLOBAL VARIABLES                                                      */
/*************************************************************************/

/* master configuration */
static glexec_config_t global_config={-1,NULL};

/* buffer space */
static int glexec_max_str_size = 128;
static char *glexec_string=NULL;

/*************************************************************************/
/* FUNCTION PROTOTYPES                                                   */
/*************************************************************************/
 
/**
 * Wrapper to mimic getc() behaviour for reading from a char buffer.
 * Needs an index, because we don't have a filepointer.
 */
int buff_getc(char* buffer, int *buff_index);

/**
 * Wrapper to mimic ungetc() behaviour for reading from a char buffer.
 * Needs an index, because we don't have a filepointer.
 */
int buff_ungetc(int c, char* buffer, int *buff_index);

/*!
 *  Initialisation of the memory area newly allocated for storing the results
 *  of parsing the glexec configuration file.
 *  \param size    Size (number of elements) of the table.
 *  \return Error code.
 */
int glexec_conf_init(int size);

/*!
 *  Parse the configuration file and return the result.
 *  \param  parse_file  File handler of the file to be parsed.
 *  \param  config      Obtained result of the parsing.
 *  \return GLEXEC_CONF_SUCCESS or failure
 */
int glexec_conf_parse_config_file(glexec_conf_parse_file_t *parse_file);

/*!
 *  Skip the remainder of the current line, i.e. all characters proceeding
 *  characters including the newline '\\n' are skipped.
 *  \param parse_file  Parse file.
 */
void glexec_conf_skip_line(glexec_conf_parse_file_t *parse_file);

/*!
 *  Get the section header in the configuration file. Section headers are
 *  enclosed by square brackets, e.g. [section]
 *  This function should be called when in the configuration file an opening
 *  bracket '[' is found. This function assumes that what follows is the
 *  section header, so it is up to the caller to make sure it is a section
 *  header. The section header returned contains all characters parsed up until
 *  the closing bracket ']'.
 *  If a newline '\\n' or eon-of-file is found an error is returned.
 *  \param parse_file  Parse file.
 *  \param section     Section header.
 *  \return Error
 */
int glexec_conf_get_target_section_id(glexec_conf_parse_file_t *parse_file,
				      char **section);

/*!
 *  When a key value pair has successfully been extracted from the configuration
 *  file, it needs to be added to the internal representation of the
 *  configuration file.
 *  \param q       New key value pair to be added.
 *  \return Error.
 */
int glexec_conf_add_key_value_pair(glexec_conf_pair_t *q);

/*!
 *  When a configuration file contained the same key twice, i.e. within
 *  a section, the old value of the key is replace by the new value.
 *  \param p  Key value pair p.
 *  \param q  Key value pair q.
 *  \return When key p and q do not match GLEXEC_CONF_SUCCESS is returned,
 *  otherwise GLEXEC_CONF_DOUBLE_KEY.
 */
int glexec_conf_check_and_replace(glexec_conf_pair_t *p,
				  const glexec_conf_pair_t * const q);

/*!
 * replaces key/value in p for this in q
 */
void glexec_conf_replace_key_value(glexec_conf_pair_t *p,
				   const glexec_conf_pair_t * const q);

/*!
 *  Get the key from the configuration file. After a section header has
 *  been found in the configuration file, it is followed by zero or more
 *  key value pairs. A key value pair looks like this:
 *   \verbatim key = value \endverbatim
 *  This functions scans the current line in the configuration file for
 *  the first appearance of the equal '=' sign. The key returned equals
 *  all characters found up until the equal sign and any trailing white
 *  space is disregarded.
 *  \param parse_file  Parse file.
 *  \param key         Returned key if found.
 *  \return Error.
 */
int glexec_conf_get_key(glexec_conf_parse_file_t *parse_file, char **key);

/*!
 *  Get the associated value of a key from the configuration file.
 *  When a key has been found, it should be followed by a value. It this
 *  value that is returned and is in effect the remainder of the current
 *  line with any leading or trailing white space removed.
 *  It is also possible to use double quotes '"', in which case all white
 *  space within these quotes is preserved.
 *  When the last none white space character on a line is the back slash
 *  '\', the next line is considered to be part of the value. The back
 *  slash is not included in the returned value.
 *  \param parse_file  Configuration file.
 *  \param value       Returned value if found.
 *  \return Error.
 */
int glexec_conf_get_value(glexec_conf_parse_file_t *parse_file, char **value);

/*!
 *  Calculate the djb2 hash from a string.
 *  \param key  String from which to calculate the hash.
 *  \return The hash number.
 */
unsigned int glexec_conf_calculate_hash(const char *key);

/************************************************************************/
/* FUNCTIONS                                                            */
/************************************************************************/

/*!
 *  Read the glexec configuration file.
 *
 *  Before reading is started it is checked if the effective user of the
 *  process has not root privileges. The config file is also checked for
 *  its permissions and ownership.
 *
 *  \param config_file  Name of the config file to read.
 *  \param config       Parsed config file.
 *
 *  \return Error.
 *
 */
int glexec_conf_read_config(const char *config_file,
			    uid_t unpriv_uid, int unpriv_gid)
{
    const char *real_config_file=config_file ? config_file : GLEXEC_CONFIG_FILE;
    glexec_conf_parse_file_t parse_file;
    int rc1;
    int rc2;
    gid_t gid=getgid(),egid=getegid();

    /* Set the defaults */
    parse_file.lineno=0;
    parse_file.index=0;
    parse_file.name=real_config_file;

    /* Read the config file */
    rc1=cgul_read_config(real_config_file, &(parse_file.buffer),
			unpriv_uid, unpriv_gid);
    /* Parse the return code */
    switch(rc1)	{
	case 0:
	    break;
	case -1:    /* I/O error */
	    glexec_log(GLEXEC_LOG_ERROR,"Failed to read configfile %s.\n",
		    real_config_file);
	    glexec_log(GLEXEC_LOG_ERROR,"  errno (%d): %s\n",
		    errno, strerror(errno));
	    return GLEXEC_CONF_COULD_NOT_OPEN_FILE;
	case -2:    /* privilege-drop error */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Could not drop priviliges when reading config file.\n");
	    return GLEXEC_CONF_NEEDS_UNPRIVILIGED_USER;
	case -3:    /* permission error (untrusted path) */
	    glexec_log(GLEXEC_LOG_ERROR,
		"Error: the config file is NOT trusted (one of its " \
		"pathcomponents is writable for others than *users* root or %s).\n",
		GLEXEC_USER);
            return GLEXEC_WRONG_PERMISSION;
	case -10:    /* permission error (unconfidential path), only when macro
			DEMAND_CONFIG_IS_CONFIDENTIAL is defined */
	    if (egid!=gid && egid!=unpriv_gid)  {
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: the config file is trusted but NOT confidential " \
		    "(it can be viewed by untrusted users and/or groups, " \
		    "i.e. other than user root or %s, and groupids %d and %d).\n",
		    GLEXEC_USER, unpriv_gid, egid);
	    } else {
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: The config file is trusted but NOT confidential " \
		    "(it can be viewed by untrusted users and/or groups, "\
		    "i.e. other than user root or %s, and groupid %d).\n",
		    GLEXEC_USER, unpriv_gid);
	    }
            return GLEXEC_WRONG_PERMISSION;
	case -4:    /* memory error */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Out of memory while reading config file.\n");
            return GLEXEC_OUT_OF_MEMORY;
	case -5:    /* unknown or safefile error */
	default:
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Error while reading config file.\n");
	    return GLEXEC_CONF_PARSING_FAILED;
    }

    /* Now parse the configuration */
    rc2=glexec_conf_parse_config_file(&parse_file);
    /* Clean up read config file buffer */
    free(parse_file.buffer); parse_file.buffer=NULL;

    /* clean up remaining memory */
    free(glexec_string); glexec_string=NULL;

    /* Check return code */
    if (rc2!=GLEXEC_SUCCESS)	{
	glexec_conf_close();
	return GLEXEC_CONF_PARSING_FAILED;
    }

    return GLEXEC_SUCCESS;
}

/*!
 * Looks up value for given key.
 * \param key
 * \return value or NULL when unset.
 */
const char *glexec_conf_value(const char *key)
{
  unsigned int slot;
  glexec_conf_pair_t *p;

  if (global_config.table_size<0) /* config uninitialized */
      return NULL;

  slot = glexec_conf_calculate_hash(key) % global_config.table_size;

  p = global_config.list[slot];

  while (p) {
    if (strcmp(key, p->key) == 0) {
      glexec_log(GLEXEC_LOG_DEBUG,
	      "Found key '%s' with value '%s'.\n", key, p->value);
      return p->value;
    }

    p = p->next;
  }

  glexec_log(GLEXEC_LOG_DEBUG, "Key '%s' was not found.\n", key);

  return NULL;
}

/*!
 *  Return the value from the configuration that is associated with the
 *  specified key. When no key is given, or its value cannot be retrieved,
 *  return the default value for this key.
 *  \param key            Search key.
 *  \param default_value  Default value for the search key.
 *  \return Value of the key in case it was found, otherwise the default
 *          value will be returned.
 *
 */
const char *glexec_conf_value_default(const char *key,const char *default_value)   {
    const char *value = NULL;

    if (!key || !(value = glexec_conf_value(key))) {
	return default_value;
    }

    return value;
}

/*!
 *  Free the resources allocated with the internal representation of
 *  the glexec config file.
 */
void glexec_conf_close(void)
{
    int i;
    glexec_conf_pair_t *p;
     
    for (i=0; i<global_config.table_size; i++)
    {   
	p = global_config.list[i];
	if (p)
	{
	    if (p->key)   free( p->key );
	    if (p->value) free( p->value );
	    if (p->next)  free( p->next );
	    free( p );
	}
    }
	
    if (global_config.list)
	free(global_config.list);

    global_config.list=NULL;
    global_config.table_size=-1;
}


/************************************************************************/
/* INTERNAL FUNCTIONS                                                   */
/************************************************************************/

/**
 * Wrapper to mimic getc() behaviour for reading from a char buffer.
 * Needs an index, because we don't have a filepointer.
 */
int buff_getc(char* buffer, int *buff_index) {
    if (buffer[*buff_index]=='\0')
	return EOF;
    else
	/* read char and move up one position */
	return (int)buffer[(*buff_index)++];
}

/**
 * Wrapper to mimic ungetc() behaviour for reading from a char buffer.
 * Needs an index, because we don't have a filepointer.
 */
int buff_ungetc(int c, char* buffer, int *buff_index) {
    /* When at start, cannot go back */
    if (*buff_index==0) return EOF;
    /* Go back on position and put in char c */
    buffer[--(*buff_index)]=(char)c;
    return c;
}

/*!
 *  Initialisation of the memory area newly allocated for storing the results
 *  of parsing the glexec configuration file.
 *  \param size    Size (number of elements) of the table.
 *  \return Error code.
 */
int glexec_conf_init(int size)
{
  /* NOTE: gcalloc never returns on error, just exits */
  global_config.list=(glexec_conf_pair_t **)gcalloc(size, sizeof(glexec_conf_pair_t)); 
  global_config.table_size = size;

  return GLEXEC_CONF_SUCCESS;
}

/*!
 *  Parse the configuration file and return the result.
 *  \param  parse_file  File handler of the file to be parsed.
 *  \param  config      Obtained result of the parsing.
 *  \return GLEXEC_CONF_SUCCESS or failure
 */
int glexec_conf_parse_config_file(glexec_conf_parse_file_t *parse_file)
{
  int c;
  int rc;
  int section_length = 0;
  int key_size;
  char *section  = NULL;
  char *key      = NULL;
  char *full_key = NULL;
  char *value    = NULL;
  glexec_conf_pair_t pair;

  if ( (rc = glexec_conf_init(CONF_ELEM_MAX)) != GLEXEC_CONF_SUCCESS )
      return rc;

  while ((c = buff_getc(parse_file->buffer,&parse_file->index)) != EOF) {
    if ((c == ' ') || (c == '\t')) {
      continue;
    }

    switch (c) {
      case '\n':
        parse_file->lineno++;
        break;
      case '#':
      case ';':
        glexec_conf_skip_line(parse_file);
        break;
      case '[':
        /* Check if there an old section stored and free it. */
        if (section) {
	  free(section);
	  section=NULL; /* Not essential, but a good safety measure */
	}
        if ( (rc = glexec_conf_get_target_section_id(parse_file, &section) )
	    != GLEXEC_CONF_SUCCESS )	{
	    glexec_conf_close();
	    return rc;
	}

        section_length = strlen(section);
        break;
      default:
        if (!section) {
          glexec_log(GLEXEC_LOG_ERROR,
	      "Missing section header for key value pairs in %s at line %d.\n",
              parse_file->name, parse_file->lineno);
	  glexec_conf_close();

          return GLEXEC_CONF_ERROR_IN_INI_FILE;
        }

        buff_ungetc(c, parse_file->buffer, &parse_file->index);

	/* First check key, if successful, also check value */
	if ( (rc = glexec_conf_get_key(parse_file, &key))==GLEXEC_CONF_SUCCESS )    {
	    rc = glexec_conf_get_value(parse_file, &value);
	    /* if value failed, free key */
	    if (rc!=GLEXEC_CONF_SUCCESS)    {
		free(key); key=NULL;
	    }
	}
	/* If either failed, bail out */
	if (rc != GLEXEC_CONF_SUCCESS)	{
	    glexec_log(GLEXEC_LOG_ERROR,
		"Problem with getting key/value pair in %s at line %d.\n",
                parse_file->name, parse_file->lineno);
	    free(section);
	    section=NULL; /* Not essential, but a good safety measure */
	    glexec_conf_close();

	    return GLEXEC_CONF_ERROR_IN_INI_FILE;
	}

        key_size = section_length + strlen(key) + 2;

	/* NOTE: gmalloc never returns on error, just exits */
        full_key = (char *)gmalloc(key_size);

        snprintf(full_key, key_size, "%s:%s", section, key);
        pair.lineno = parse_file->lineno;
        pair.key    = full_key;
        pair.value  = value;
        if ((rc=glexec_conf_add_key_value_pair(&pair))
		!=GLEXEC_CONF_SUCCESS) {
	  free(full_key);
	  free(key); free(value);
	  free(section);
          section=NULL;
	  glexec_conf_close();
	  return rc;
	}


        /* Clean up allocated strings. */
        free(key); key=NULL;
        free(value); value=NULL;
        free(full_key); full_key=NULL;
    }
  }

  free(section);
  section=NULL; /* Not essential, but a good safety measure */

  return GLEXEC_CONF_SUCCESS;
}


/*!
 *  Skip the remainder of the current line, i.e. all characters proceeding
 *  characters including the newline '\\n' are skipped.
 *
 *  \param parse_file  Parse file.
 */
void glexec_conf_skip_line(glexec_conf_parse_file_t *parse_file)
{
  int c;

  do {
    c = buff_getc(parse_file->buffer,&parse_file->index);
  } while (c != '\n' && c!=EOF); /* Important to check for EOF !! */

  parse_file->lineno++;
}


/*!
 *  Get the section header in the configuration file. Section headers are
 *  enclosed by square brackets, e.g. [section]
 *  This function should be called when in the configuration file an opening
 *  bracket '[' is found. This function assumes that what follows is the
 *  section header, so it is up to the caller to make sure it is a section
 *  header. The section header returned contains all characters parsed up until
 *  the closing bracket ']'.
 *  If a newline '\\n' or eon-of-file is found an error is returned.
 *  \param parse_file  Parse file.
 *  \param section     Section header.
 *  \return Error
 */
int glexec_conf_get_target_section_id(glexec_conf_parse_file_t *parse_file,
				      char **section)
{
  int c, n;

  n = 0;

  if (!glexec_string)
	/* NOTE: gmalloc never returns on error, just exits */
	glexec_string = (char *)gmalloc(glexec_max_str_size);

  do {
    c = buff_getc(parse_file->buffer,&parse_file->index);
    if ((c == '\n') || (c == EOF)) {
      parse_file->lineno++;
      return GLEXEC_CONF_LABEL_NOT_CLOSED;
    }

    if (c == ']') {
      break;
    }

    /* Store the new character at the end of the string and increase the
     * total number of characters and ensure it does not exceed the maximum
     * string length. If it does, reallocate twice as much memory then
     * preciously allocated. */
    glexec_string[n++] = (char)c;

    if (n >= glexec_max_str_size) {
      glexec_max_str_size *= 2;
      /* NOTE: grealloc never returns on error, just exits */
      glexec_string=(char *)grealloc(glexec_string, glexec_max_str_size);
    }

  } while (c != ']');

  /* Make sure the string is \0 terminated. */
  glexec_string[n] = '\0';

  /* NOTE: gstrdup never returns on error, just exits */
  *section = gstrdup(glexec_string);

  return GLEXEC_CONF_SUCCESS;
}


/*!
 *  When a key value pair has successfully been extracted from the configuration
 *  file, it needs to be added to the internal representation of the
 *  configuration file.
 *  \param q       New key value pair to be added.
 *  \return Error.
 */
int glexec_conf_add_key_value_pair( glexec_conf_pair_t *q)
{
  unsigned int slot;
  glexec_conf_pair_t *p;
  int rc;

  if (global_config.table_size<0) /* config uninitialized */
      return GLEXEC_CONF_UNINITIALIZED;

  slot = glexec_conf_calculate_hash(q->key) % global_config.table_size;

  p = *(global_config.list + slot);

  /* Search for the last element of the list. */
  if (p) {
    if ((rc=glexec_conf_check_and_replace(p, q))!=GLEXEC_CONF_SUCCESS) {
      return rc;
    }

    while (p->next) {
      if ((rc=glexec_conf_check_and_replace(p, q))!=GLEXEC_CONF_SUCCESS) {
        return rc;
      }

      p = p->next;
    }
    /* NOTE: gmalloc never returns on error, just exits */
    p->next = (glexec_conf_pair_t *)gmalloc(sizeof(glexec_conf_pair_t));
    p = p->next;
  } else {
    /* NOTE: gmalloc never returns on error, just exits */
    *(global_config.list + slot) = p = (glexec_conf_pair_t *)gmalloc(sizeof(glexec_conf_pair_t));
  }

  if (!p)
      return GLEXEC_CONF_OUT_OF_MEMORY;

  p->lineno = q->lineno;
  /* NOTE: gstrdup never returns on error, just exits */
  p->key    = gstrdup(q->key);
  p->value  = gstrdup(q->value);
  p->next   = NULL;

  return GLEXEC_CONF_SUCCESS;
}


/*!
 *  When a configuration file contained the same key twice, i.e. within
 *  a section, the old value of the key is replace by the new value.
 *  \param p  Key value pair p.
 *  \param q  Key value pair q.
 *  \return When key p and q do not match GLEXEC_CONF_SUCCESS is returned,
 *  otherwise GLEXEC_CONF_DOUBLE_KEY.
 */
int glexec_conf_check_and_replace(glexec_conf_pair_t *p,
				  const glexec_conf_pair_t * const q)
{
  /* Check if the same key exists. */
  if (strcmp(p->key, q->key) == 0) {
    glexec_log(GLEXEC_LOG_WARNING,
	    "Warning: key defined twice at lines %d and %d! Previous key and value are disregarded.\n",
            p->lineno, q->lineno);
    glexec_log(GLEXEC_LOG_DEBUG, "         key       = '%s'\n", p->key);
    glexec_log(GLEXEC_LOG_DEBUG, "         old value = '%s'\n", p->value);
    glexec_log(GLEXEC_LOG_DEBUG, "         new value = '%s'\n", q->value);
    glexec_conf_replace_key_value(p, q);
    /* Return a warning that a double key has been found. */
/*    return GLEXEC_CONF_DOUBLE_KEY; -> should still succeed -> only warning */
  }

  return GLEXEC_CONF_SUCCESS;
}


/*!
 * replaces key/value in p for this in q
 */
void glexec_conf_replace_key_value(glexec_conf_pair_t *p,
				   const glexec_conf_pair_t * const q)
{
  /* Throw away any old values... */
  free(p->key);
  free(p->value);

  /* ...and replace them with new ones. */
  p->lineno = q->lineno;
  /* NOTE: gstrdup never returns on error, just exits */
  p->key    = gstrdup(q->key);
  p->value  = gstrdup(q->value);
}

/*!
 *  Get the key from the configuration file. After a section header has
 *  been found in the configuration file, it is followed by zero or more
 *  key value pairs. A key value pair looks like this:
 *   \verbatim key = value \endverbatim
 *  This functions scans the current line in the configuration file for
 *  the first appearance of the equal '=' sign. The key returned equals
 *  all characters found up until the equal sign and any trailing white
 *  space is disregarded.
 *  \param parse_file  Parse file.
 *  \param key         Returned key if found.
 *  \return Error.
 */
int glexec_conf_get_key(glexec_conf_parse_file_t *parse_file, char **key)
{
  int c, n;

  if (!glexec_string) {
      /* NOTE: gmalloc never returns on error, just exits */
      glexec_string = (char *)gmalloc(glexec_max_str_size);
  }

  for (n=0;; n++ ) {
    c = buff_getc(parse_file->buffer,&parse_file->index);

    if ((c == '\n') || (c == EOF)) {
      parse_file->lineno++;

      return GLEXEC_CONF_PARSING_FAILED;
    }

    /* Check if we have found the end of the key. */
    if (c == '=') {
      --n;   /* Correct the counter as the value has not been stored. */
      break;
    }

    if (n >= glexec_max_str_size) {
      glexec_max_str_size *= 2;
      /* NOTE: grealloc never returns on error, just exits */
      glexec_string=(char *)grealloc(glexec_string,glexec_max_str_size);
    }

    glexec_string[n] = (char)c;
  }

  /* Remove any trailing blanks. */
  while ((glexec_string[n] == ' ') || (glexec_string[n] == '\t')) {
    glexec_string[n--] = '\0';
  }

  /* NOTE: gstrdup never returns on error, just exits */
  *key = gstrdup(glexec_string);

  return GLEXEC_CONF_SUCCESS;
}

/*!
 *  Get the associated value of a key from the configuration file.
 *  When a key has been found, it should be followed by a value. It this
 *  value that is returned and is in effect the remainder of the current
 *  line with any leading or trailing white space removed.
 *  It is also possible to use double quotes '"', in which case all white
 *  space within these quotes is preserved.
 *  When the last none white space character on a line is the back slash
 *  '\', the next line is considered to be part of the value. The back
 *  slash is not included in the returned value.
 *  \param parse_file  Configuration file.
 *  \param value       Returned value if found.
 *  \return Error.
 */
int glexec_conf_get_value(glexec_conf_parse_file_t *parse_file, char **value)
{
  int c, n;
  int quoted = 0;
  int continuation = 0;

  if (global_config.table_size<0) /* config uninitialized */
      return GLEXEC_CONF_UNINITIALIZED;

  if (!glexec_string)
      /* NOTE: gmalloc never returns on error, just exits */
      glexec_string = (char *)gmalloc(glexec_max_str_size);

  /* Remove any leading white spaces. */
  do {
    c = buff_getc(parse_file->buffer,&parse_file->index);
  } while ((c == ' ') || (c == '\t'));

  buff_ungetc(c, parse_file->buffer, &parse_file->index);

  for (n=0;;) {
    c = buff_getc(parse_file->buffer, &parse_file->index);

    if (c == '\\') {
      continuation = 1;
      continue;
    }

    if ((c == '\n') || (c == EOF)) {
      parse_file->lineno++;

      if (quoted == 1) {
        if (continuation == 1) {
          continuation = 0;
          continue;
        }
        return GLEXEC_CONF_ERROR_IN_INI_FILE;
      }
      break;
    }

    if (continuation == 1) {
      return GLEXEC_CONF_ERROR_IN_INI_FILE;
    }

    if (c == '"') {
      if (quoted) {
        quoted = 2;
        break;
      } else {
        quoted = 1;
        continue;
      }
    }

    glexec_string[n++] = (char)c;

    if (n >=glexec_max_str_size) {
      glexec_max_str_size *= 2;
      /* NOTE: grealloc never returns on error, just exits */
      glexec_string= (char *)grealloc(glexec_string, glexec_max_str_size);
    }
  }

  /* Make sure the string is '\0' terminated. */
  glexec_string[n] = '\0';

  --n;

  while ( (n > 0) && (glexec_string[n] == ' ')) {
    glexec_string[n--] = '\0';
  }

  /* NOTE: gstrdup never returns on error, just exits */
  *value = gstrdup(glexec_string);

  return GLEXEC_CONF_SUCCESS;
}

/*!
 *  Calculate the djb2 hash from a string.
 *  \param key  String from which to calculate the hash.
 *  \return The hash number.
 */
unsigned int glexec_conf_calculate_hash(const char *key)
{
  unsigned long hash = 5381;
  int c;

  while ((c = *key++)) {
    hash = ((hash << 5) + hash) + c;
  }

  return hash;
}


