/**
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors:
 * 2009-
 *	Oscar Koeroo <okoeroo@nikhef.nl>
 *	Mischa Sall\'e <msalle@nikhef.nl>
 *	NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *	Gerben Venekamp <venekamp@nikhef.nl>
 *	NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 *  \file glexec_environ.h
 *
 *  Provides environment and current working directory related functions
 */

#ifndef _GLEXEC_ENVIRON_H
#define _GLEXEC_ENVIRON_H

/* For GLEXEC_ERROR */
#include "glexec_exit.h"

/* For env_t */
#include "environ/environ.h"

/************************************************************************/
/* DEFINES	                                                        */
/************************************************************************/

/* CWD used during the gLExec run */
#ifndef GLEXEC_RUNTIME_CWD
#define GLEXEC_RUNTIME_CWD  "/"
#endif

/************************************************************************/
/* FUNCTION PROTOTYPES                                                  */
/************************************************************************/

/*!
 * Frees the internal static env_t variables (backup of original env and
 * target env).
 */
void glexec_free_env(void);

/*!
 * Getter function for the internal target environment
 * \return target environment
 */
env_t glexec_get_target_env(void);

/*!
 * Getter function for the internal caller current working directory
 * \return caller current working directory
 */
char *glexec_get_caller_cwd(void);

/*!
 * Sets the glexec runtime current working directory to the safe
 * GLEXEC_RUNTIME_CWD instead of the caller's working directory. The original
 * working directory is saved and used for finding the input proxies.
 * \return new current working directory
 */
const char *glexec_set_runtime_cwd(void);

/*!
 * Prepends value of specified variable with cwd of calling user.
 * \param var should contain
 * \param value optionally the (old) value of the variable, if NULL, it will be
 * obtained dynamically.
 * \param *newvalue new value, when different from value it points to a static
 * buffer overwritten at next call, or NULL.
 * \return 0 on success, -1 on error, in which case errno is set.
 */
int glexec_make_var_absolute(const char *var,char *value,char **newvalue);

/*!
 * Makes a backup of the original environment, then completely clean the
 * internal gLExec environment except for those in GLEXEC_ENV_WHITELIST or
 * matched bt GLEXEC_ENV_PATTERN. It also does some basic sanity checks on the
 * result. This function should be called as soon as possible.
 * \return original environment
 */
env_t glexec_clean_environment(void);

/*!
 * Sets up a clean new environment for the target environment from the original
 * backup using the whitelist as passed in. NOTE that this variable will be
 * freed.
 * In addition those variables matching GLEXEC_ENV_PATTERN are added.
 * \param env_whitelist variable whitelist.
 * \return target environment
 */
env_t glexec_setup_target_environment(env_t *env_whitelist);

/*!
 *  Set an environment variable to the specified value, only if the value found
 *  matches the regular expression
 *  \param var    Environment variable that needs to be settings
 *  \param value  The new value for the environment variable
 *  \param regex  Regular expression that defines valid values.
 *  \return value set or NULL when nothing was set. Note that when the regex
 *  itself is invalid (not when it doesn't match), the value will still be set.
 */
const char *glexec_setenv_regex(const char *var,const char *value,const char *regex);

/*!
 * Sets a name - value pair into the target environment
 * \param var variable
 * \param value
 * \return (new) target_environment
 */
env_t glexec_target_setenv(const char *var, const char *value);

/*!
 * Unsets a var from the target environment
 * \param var variable
 * \return (new) target_environment
 */
env_t glexec_target_unsetenv(const char *var);
#endif
