/**
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 * glexec_exit.c -- exit handling functions for gLExec
 * \file glexec_error.c
 */

#include "glexec_exit.h"

#include "glexec.h"
#include "glexec_log.h"
#include "glexec_conf.h"
#include "glexec_account.h"
#include "glexec_environ.h"
#include "glexec_util.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <pwd.h>
#include <errno.h>

/************************************************************************/
/* DEFINES AND TYPES                                                    */
/************************************************************************/
#define ERR_MAX_BUF 1024

typedef enum GLEXEC_LOG_TO_WHO_E
{
    LOG_TO_ADMIN          = 1,
    LOG_TO_USER           = 2,
    LOG_TO_ADMIN_AND_USER = 3
} GLEXEC_LOG_TO_WHO_T;


/*************************************************************************/
/* FUNCTION PROTOTYPES                                                   */
/*************************************************************************/
/*!
 *  Display error messages based on the error code.
 *  \param rc  Error code.
 *  \param forwho
 */
void display_error(int rc, GLEXEC_LOG_TO_WHO_T forwho);

/*!
 *  Display and log a multiline error message.
 *  \param print_err_fmt NULL terminated list of error lines.
 */
void display_and_log_error_array(char *print_err_fmt[]);

/*************************************************************************/
/* FUNCTION DECLARATIONS                                                 */
/*************************************************************************/

/*!
 *  Exit glexec and free allocated resources.
 *  \param glexec_internal_rc	Error code leading to the exit.
 *  \param child_return_code	Error code of the child process
 */
void glexec_exit(int glexec_internal_rc, int child_return_code)
{
    glexec_conf_close();
    glexec_free_cached_passwds();
    glexec_free_cached_groups();
    glexec_free_env();

    /*
     * Decision from Maarten Litmaath for clean and clear error codes
     *
     * 201 - client error
     * 202 - internal error
     * 203 - authz error
     * 
     * To deal with the application using any of those codes one would add this
     * case:
     * 
     * 204 - application exited with code 201, 202, 203 or 204
     */


    /*
     * When internal error code returns GLEXEC_SUCCESS and when the 
     * child_return_code returns GLEXEC_NO_CHILD_EXIT_CODE then this indicates
     * the master gLExec to have exited nicely
     */
    if ((glexec_internal_rc == GLEXEC_SUCCESS) && 
        (child_return_code == GLEXEC_NO_CHILD_EXIT_CODE))
    {
        glexec_close_log();
        exit (0);
    }


    /*
     * When internal error code returns GLEXEC_SUCCESS and when the 
     * child_return_code returns something else, then this is the return code 
     * of the child GLEXEC_NO_CHILD_EXIT_CODE then this indicates the return 
     * of the child gLExec
     */
    if (glexec_internal_rc == GLEXEC_SUCCESS)
    {
        /* Child return code overlap detection */
        switch (child_return_code)
        {
            /* 201, 202, 203, 204 are the error codes used for gLExec */
            case GLEXEC_EXIT_USER_ERROR               :
            case GLEXEC_EXIT_INTERNAL_ERROR           :
            case GLEXEC_EXIT_AUTHZ_ERROR              :
            case GLEXEC_EXIT_CHILD_PROCESS_RC_OVERLAP :
                glexec_close_log();
		/* 204 - application exited with code 201, 202, 203 or 204 */
                exit (GLEXEC_EXIT_CHILD_PROCESS_RC_OVERLAP);
            default :
                glexec_close_log();
                exit (child_return_code);
        }
    }
    else
    {
	/* Switching on the internal gLExec error codes - Everything is an error
	 * from here. */
        switch (glexec_internal_rc)
        {
            case  GLEXEC_NO_SOURCE_PROXY              :
            case  GLEXEC_SETENV_ERROR                 :
            case  GLEXEC_STAT_ERROR                   :
            case  GLEXEC_STAT_ERROR_EXECUTABLE_CMD    :
            case  GLEXEC_STAT_ERROR_CMD_SYMLINKED     :
            case  GLEXEC_STAT_ERROR_CURRENT_WORK_DIR  :
            case  GLEXEC_CHPWD_ERROR                 :
            case  GLEXEC_STAT_CMD_PERMS_CANT_EXEC     :
            case  GLEXEC_COPY_PROXY_FAILED            :
            case  GLEXEC_PROXY_IS_LINK                :
            case  GLEXEC_PROXY_OPEN_FAILED            :
            case  GLEXEC_SOURCE_PROXY_OPEN_FAILED     :
            case  GLEXEC_PROXY_CLOSE_FAILED           :
            case  GLEXEC_FAILURE_WRITING_PROXY        :
            case  GLEXEC_WRONG_OWNERSHIP              :
            case  GLEXEC_PROG_WRONG_OWNERSHIP              :
            case  GLEXEC_PROXY_TOO_LARGE              :
            case  GLEXEC_PRIVILEGED_USER              :
            case  GLEXEC_FILE_CHANGED                 :
            case  GLEXEC_FAILURE_READING_PROXY_ENV    :
            case  GLEXEC_FAILURE_READING_PROXY_FILE   :
            case  GLEXEC_WRONG_SET_OF_ARGS            :
            case  GLEXEC_NO_CHILD_EXIT_CODE           :
            case  GLEXEC_WRONG_PERMISSION_PROXY       :
            case  GLEXEC_WRONG_PERMISSION_ON_CMD      :
	    case  GLEXEC_LOCK_FAILURE                 :
	    case  GLEXEC_USER_TRIED_V		      :
                display_error (glexec_internal_rc, LOG_TO_ADMIN_AND_USER);
		glexec_close_log();
		/* 201 - client error */
                exit (GLEXEC_EXIT_USER_ERROR);

	    case  GLEXEC_ENV_CLEAN_ERROR	      :
	    case  GLEXEC_TARGET_ENV_ERROR	      :
            case  GLEXEC_LCAS_INIT_FAIL               :
            case  GLEXEC_LCMAPS_INIT_FAIL             :
            case  GLEXEC_INCOMPLETE_MAPPING_RESULT    :
            case  GLEXEC_BACKLOG_FAILED               :
            case  GLEXEC_CONF_VALUE_MALFORMED         :
            case  GLEXEC_CONF_FAIL_USER_ID_SWITCH_BY  :
            case  GLEXEC_WRONG_PERMISSION             :
            case  GLEXEC_DIR_WRONG_OWNERSHIP          :
            case  GLEXEC_RSL_FAILED                   :
            case  GLEXEC_OUT_OF_MEMORY                :
            case  GLEXEC_ALLOC_ERROR                  :
            case  GLEXEC_COULD_NOT_DROP_PRIVILEGES    :
            case  GLEXEC_DID_NOT_USERSWITCH	      :
	    case  GLEXEC_STILL_EFFECTIVE_ROOT	      :
            case  GLEXEC_WRONG_USERSWITCH	      :
            case  GLEXEC_DOUBLE_USERSWITCH	      :
	    case  GLEXEC_TOO_MANY_GROUPS	      :
            case  GLEXEC_WRONG_NUMBER_OF_BYTES_READ   :
            case  GLEXEC_REJECTED_ENV_VAR             :
            case  GLEXEC_ERROR_CHDIR		      :
            case  GLEXEC_UNKNOWN_USER                 :
            case  GLEXEC_UNKNOWN_GROUP                :
            case  GLEXEC_FAILED_TO_RESTORE_EFF_UID    :
            case  GLEXEC_MAPPED_TO_FORBIDDEN_USER     :
            case  GLEXEC_MAPPED_TO_FORBIDDEN_GROUP    :
            case  GLEXEC_MEM_ALLOC_FAILURE            :
            case  GLEXEC_INSUFFICIENT_PRIVILEGES      :
            case  GLEXEC_UNCLASIFIED                  :
            case  GLEXEC_LOG_FILE_CLOSE_FAIL          :
                display_error (glexec_internal_rc, LOG_TO_ADMIN);
                glexec_close_log();
		/* 202 - internal error */
                exit (GLEXEC_EXIT_INTERNAL_ERROR);
                
            case  GLEXEC_LOG_FILE_OPEN_FAIL           :
            case  GLEXEC_LOG_FILE_WRONG_PERMISSION    :
            case  GLEXEC_LCAS_FAILED                  :
            case  GLEXEC_LCMAPS_FAILED                :
            case  GLEXEC_USER_NOT_WHITELISTED         :
                display_error (glexec_internal_rc, LOG_TO_ADMIN_AND_USER);
                glexec_close_log();
		/* 203 - authz error */
                exit (GLEXEC_EXIT_AUTHZ_ERROR);

            default :
                /* Undefined error, thus internal error */
                glexec_close_log();
		/* 202 - undefined internal glexec_exit */
                exit (GLEXEC_EXIT_INTERNAL_ERROR);
        }
    }
}

/*!
 * Called when the execve fails: prints proper error message to stdout and
 * glexec log.
 * \param cmd command which was tried.
 * \param cwd current working directory, ignored when cmd is absolute.
 */
void glexec_execve_exit(char *cwd, char *fullcmd, char* realcmd)	{
    int len;
    char buffer[GLEXEC_MAX_LOG_LINE];
    char *print_err_fmt[6]={NULL,NULL,NULL,NULL,NULL,NULL}; /* list of NULL terminated char* pointers */

    glexec_log(GLEXEC_LOG_ERROR,
	    "ERROR: execve of %s (=%s) in %s failed: %s\n",
	    fullcmd,realcmd,cwd,strerror(errno));
    switch (errno) {
	case ENOTDIR      :   
	    print_err_fmt[0]="A component of the path prefix is not a directory.\n";
	    display_and_log_error_array(print_err_fmt);
	    break;
	case ENAMETOOLONG :
	    print_err_fmt[0]="A component of a pathname exceeded {NAME_MAX} characters, or an entire path name exceeded {PATH_MAX} characters.\n";
	    display_and_log_error_array(print_err_fmt);
	    break;
	case ENOENT       :  
	    print_err_fmt[0]="The new process file does not exist.\n";
	    display_and_log_error_array(print_err_fmt);
	    break;
	case ELOOP        : 
	    print_err_fmt[0]="Too many symbolic links were encountered in translating the pathname.\n";
	    display_and_log_error_array(print_err_fmt);
	    break;
	case EACCES       : 
	    print_err_fmt[0]="Search permission is denied for a component of the path prefix,\n";
	    print_err_fmt[1]="  or the new process file is not an ordinary file,\n";
	    print_err_fmt[2]="  or the new process file mode denies execute permission,\n";
	    print_err_fmt[3]="  or the new process file is on a filesystem mounted with execution disabled (MNT_NOEXEC in <sys/mount.h>).\n";
	    display_and_log_error_array(print_err_fmt);
	    break;
	case ENOEXEC      :   
	    print_err_fmt[0]="The new process file has the appropriate access permission, but has an invalid magic number in its header.\n";
	    display_and_log_error_array(print_err_fmt);
	    break;
	case ETXTBSY      :   
	    print_err_fmt[0]="The new process file is a pure procedure (shared text) file that is currently open for writing or reading by some process.\n";
	    display_and_log_error_array(print_err_fmt);
	    break;
	case ENOMEM       :  
	    print_err_fmt[0]="The new process requires more virtual memory than is allowed by the imposed maximum (getrlimit(2)).\n";
	    display_and_log_error_array(print_err_fmt);
	    break;
	case E2BIG        :             
	    print_err_fmt[0]="The number of bytes in the new process's argument list is larger than the system-imposed limit.  This limit is specified by the sysctl(3) MIB variable KERN_ARGMAX.\n";
	    display_and_log_error_array(print_err_fmt);
	    break;
	case EFAULT       :  
	    print_err_fmt[0]="The new process file is not as long as indicated by the size values in its header.\n";
	    print_err_fmt[1]="  or path, argv, or envp point to an illegal address.\n";
	    display_and_log_error_array(print_err_fmt);
	    break;
	case EIO         :  
	    print_err_fmt[0]="An I/O error occurred while reading from the file system.\n";
	    display_and_log_error_array(print_err_fmt);
	    break;
	default	     :
	    print_err_fmt[0]="An unknown error occurred during execve().\n";
	    display_and_log_error_array(print_err_fmt);
	    break;
    }
    len=snprintf(buffer,GLEXEC_MAX_LOG_LINE,
		"Command to be executed: \"%s\" (=\"%s\").\n",fullcmd,realcmd);
    print_err_fmt[0]=buffer;
    if (len>=GLEXEC_MAX_LOG_LINE)
	sprintf(&(print_err_fmt[0][GLEXEC_MAX_LOG_LINE-5]),"...\n");
    print_err_fmt[1]=NULL;
    display_and_log_error_array(print_err_fmt);

    /* Now still have to close the logfile... */
    glexec_log(GLEXEC_LOG_DEBUG,"Closing the logfile (after execve has failed).\n");
    glexec_close_log();
    /*
     * Tested with bash, 
     * when an error is triggered, 
     * the execve() passed program returns the exit code of '126' when the
     * command can't be executed.
     * Other errors might also be triggered by it now.
     */
    exit(GLEXEC_EXECVE_FAILED_EXIT_CODE);
}

/*************************************************************************/
/* INTERNAL FUNCTIONS		                                         */
/*************************************************************************/

/*!
 *  Display error messages based on the error code.
 *  \param rc  Error code.
 *  \param forwho
 */
void display_error(int rc, GLEXEC_LOG_TO_WHO_T forwho) 
{
    char *envval1,*envval2,*envval3;
    char error_msg_buf[ERR_MAX_BUF];
    memset(error_msg_buf, 0, ERR_MAX_BUF);
    uid_t uid;
    struct passwd *pw;

    /* Switching on the internal gLExec error codes - Everything is an error
     * from here. */
    switch (rc)
    {
        /* User error */
        case  GLEXEC_WRONG_SET_OF_ARGS            :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  The wrong number of arguments was specified on the command line. Use '%s -h' for more information\n", 
		    GLEXEC_PACKAGE_NAME);
            break;
        case  GLEXEC_NO_SOURCE_PROXY              :
	    envval1=getenv("GLEXEC_SOURCE_PROXY");
            snprintf (error_msg_buf, ERR_MAX_BUF, "  The source proxy was not found at '%s'.\n", envval1 ? envval1 : "(unset variable GLEXEC_SOURCE_PROXY)");
            break;
        case  GLEXEC_SETENV_ERROR              :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Error while trying to set environ ment variable.\n");
            break;
        case  GLEXEC_STAT_ERROR                   :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  The stat syscall returned with an error (not clear for what, please file a bug).\n");
            break;
        case  GLEXEC_STAT_ERROR_EXECUTABLE_CMD    :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Error while examining the executable that is about to be executed.\n");
            break;
        case  GLEXEC_STAT_ERROR_CMD_SYMLINKED     :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  The program to execute is symlinked.\n");
            break;
        case  GLEXEC_STAT_ERROR_CURRENT_WORK_DIR  :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  The stat syscall returned with an error for the current working directory.\n");
            break;
        case  GLEXEC_CHPWD_ERROR  :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Cannot do a chdir to the target user directory.\n");
            break;
        case  GLEXEC_STAT_CMD_PERMS_CANT_EXEC     :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  The program to execute is either not executable or the (mapped) account doesn't have sufficient file permissions to execute it.\n");
            break;
        case  GLEXEC_COPY_PROXY_FAILED            :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Copying of the proxy failed.\n");
            break;
        case  GLEXEC_WRONG_PERMISSION_ON_CMD      :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  The permissions on the executable are wrong. It must not be writable for the world (could be a dismissed statement), no setuid or setgid and must be executable for the (mapped) user.\n");
            break;
        case  GLEXEC_WRONG_PERMISSION_PROXY:
            snprintf (error_msg_buf, ERR_MAX_BUF, "  The permissions on $GLEXEC_SOURCE_PROXY and/or $GLEXEC_CLIENT_CERT are wrong: they may NOT be readable for anyone but the owner.\n");
            break;
        case  GLEXEC_PROXY_IS_LINK         :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  The path to $GLEXEC_SOURCE_PROXY or $GLEXEC_CLIENT_CERT seems to be a symlink. Must use a path to the actual file.\n");
            break;
        case  GLEXEC_PROXY_OPEN_FAILED     :
	    envval1=getenv("GLEXEC_CLIENT_CERT");
	    envval2=getenv("GLEXEC_SOURCE_PROXY");
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Failed to open $GLEXEC_CLIENT_CERT=%s or $GLEXEC_SOURCE_PROXY=%s.\n",
		    (envval1 ? envval1 : "(NULL)"),
		    (envval2 ? envval2 : "(NULL)"));
            break;
        case  GLEXEC_SOURCE_PROXY_OPEN_FAILED     :
	    envval1=getenv("GLEXEC_SOURCE_PROXY");
            snprintf(error_msg_buf, ERR_MAX_BUF, "  Failed to open $GLEXEC_SOURCE_PROXY=%s.\n",
		    (envval1 ? envval1 : "(NULL)"));
            break;
        case  GLEXEC_PROXY_CLOSE_FAILED           :
	    envval1=getenv("GLEXEC_CLIENT_CERT");
	    envval2=getenv("GLEXEC_SOURCE_PROXY");
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Failed to close $GLEXEC_CLIENT_CERT=%s or $GLEXEC_SOURCE_PROXY=%s.\n",
		    (envval1 ? envval1 : "(NULL)"),
		    (envval2 ? envval2 : "(NULL)"));
            break;
        case  GLEXEC_FAILURE_WRITING_PROXY        :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Failure writing proxy file\n");
            break;
        case  GLEXEC_WRONG_OWNERSHIP              :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Expecting a different ownership for directory or file.\n");
            break;
        case  GLEXEC_DIR_WRONG_OWNERSHIP              :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Expecting a different ownership for the directory.\n");
            break;
        case  GLEXEC_PROG_WRONG_OWNERSHIP              :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Expecting a different ownership for the program.\n");
            break;
        case  GLEXEC_PROXY_TOO_LARGE       :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  The $GLEXEC_CLIENT_CERT or $GLEXEC_SOURCE_PROXY is too large\n");
            break;
        case  GLEXEC_PRIVILEGED_USER              :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  (internal problem) gLExec is still running with privileged credentials where they should have been dropped.\n");
            break;
        case  GLEXEC_FILE_CHANGED                 :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  gLExec has detected an input file change during the use of the file. It's unknown if this file-jacking was accidental or intentional.\n");
            break;
        case  GLEXEC_FAILURE_READING_PROXY_ENV    :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  environment variable $GLEXEC_CLIENT_CERT is empty.\n");
            break;
        case  GLEXEC_FAILURE_READING_PROXY_FILE   :
	    envval1=getenv("GLEXEC_CLIENT_CERT");
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Unable to read the proxy file from $GLEXEC_CLIENT_CERT=%s.\n",
			(envval1 ? envval1 : "(NULL)"));
            break;
        case  GLEXEC_LOCK_FAILURE                 :
	    envval1=getenv("GLEXEC_CLIENT_CERT");
	    envval2=getenv("GLEXEC_SOURCE_PROXY");
	    envval3=getenv("GLEXEC_TARGET_PROXY");
	    if (envval3!=NULL)
		snprintf (error_msg_buf, ERR_MAX_BUF, "  Failed to lock $GLEXEC_CLIENT_CERT=%s, $GLEXEC_SOURCE_PROXY=%s or $GLEXEC_TARGET_PROXY=%s.\n",
			(envval1 ? envval1 : "(NULL)"),
			(envval2 ? envval2 : "(NULL)"),
			envval3);
	    else
		snprintf (error_msg_buf, ERR_MAX_BUF, "  Failed to lock $GLEXEC_CLIENT_CERT=%s, $GLEXEC_SOURCE_PROXY=%s or destination proxy.\n",
			(envval1 ? envval1 : "(NULL)"),
			(envval2 ? envval2 : "(NULL)"));
            break;
	case  GLEXEC_USER_TRIED_V		  :
	    snprintf (error_msg_buf, ERR_MAX_BUF, "  User is not permitted to use -V option.\n");
	    break;
        /* Administrative / internal failure domain */
        case  GLEXEC_CONF_FAIL_LOCK_MECH          :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  The value for one of the locking options \"glexec:*lock_mechanism\" is unexpected.\n");
            break;
        case  GLEXEC_LCAS_INIT_FAIL               :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Initialisation of LCAS failed.\n");
            break;
        case  GLEXEC_LCMAPS_INIT_FAIL             :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Initialisation of LCMAPS failed.\n");
            break;
        case  GLEXEC_INCOMPLETE_MAPPING_RESULT    :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Incomplete mapping result returned (could be missing user or group information).\n");
            break;
        case  GLEXEC_BACKLOG_FAILED               :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  The backlog function failed to create a backlog entry.\n");
            break;
        case  GLEXEC_CONF_FAIL_USER_ID_SWITCH_BY  :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  The value for the option \"user_identity_switch_by\" is unexpected.\n");
            break;
        case  GLEXEC_CONF_VALUE_MALFORMED         :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Configuration file contains values that are not understood.\n");
            break;
        case  GLEXEC_WRONG_PERMISSION             :
            /* To do, must be split to be specific to the file that has wrong permission bits */
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Wrong permissions on the executable that is to be executed or source proxy\n");
            break;
        case  GLEXEC_RSL_FAILED                   :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Generating the RSL string failed.\n");
            break;
        case  GLEXEC_OUT_OF_MEMORY                :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Out of memory\n");
            break;
        case  GLEXEC_ALLOC_ERROR                 :
	    snprintf (error_msg_buf, ERR_MAX_BUF, "  Memory allocation error (not out-of-memory)\n");
            break;
        case  GLEXEC_COULD_NOT_DROP_PRIVILEGES    :
	    snprintf (error_msg_buf, ERR_MAX_BUF,
		    "  Couldn't drop privileges, gLExec doesn't have sufficient privileges to drop.\n");
            break;
        case  GLEXEC_DID_NOT_USERSWITCH		  :
	    snprintf (error_msg_buf, ERR_MAX_BUF,
		    "  LCMAPS should have changed identity, but lcmaps_posix_enf did not run.\n");
            break;
        case  GLEXEC_STILL_EFFECTIVE_ROOT	  :
	    snprintf (error_msg_buf, ERR_MAX_BUF,
		    "  Identity has changed, but without dropping effective root bit.\n");
            break;
        case  GLEXEC_WRONG_USERSWITCH		  :
	    snprintf (error_msg_buf, ERR_MAX_BUF, "  Identity has been changed to wrong account.\n");
            break;
        case  GLEXEC_DOUBLE_USERSWITCH		  :
	    snprintf (error_msg_buf, ERR_MAX_BUF, "  gLExec is configured to do the userswitch, but LCMAPS already changed identity.\n");
            break;
	case  GLEXEC_TOO_MANY_GROUPS		  :
	    snprintf (error_msg_buf, ERR_MAX_BUF, "  Too many groups for target user.\n");
	    break;
        case  GLEXEC_WRONG_NUMBER_OF_BYTES_READ   :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Wrong number of bytes read.\n");
            break;
        case  GLEXEC_REJECTED_ENV_VAR             :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Rejected environment variables.\n");
            break;
	case  GLEXEC_ERROR_CHDIR		  :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Chdir failed.\n");
            break;
        case  GLEXEC_UNKNOWN_USER                 :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Unknown Unix user on this system\n");
            break;
        case  GLEXEC_UNKNOWN_GROUP                :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Unknown Unix group on this system\n");
            break;
        case  GLEXEC_FAILED_TO_RESTORE_EFF_UID    :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Failed to restore effective UID in the process\n");
            break;
        case  GLEXEC_MAPPED_TO_FORBIDDEN_USER     :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Failed mapping, mapped to forbidden userid (like a system account)\n");
            break;
        case  GLEXEC_MAPPED_TO_FORBIDDEN_GROUP    :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Failed mapping, mapped to forbidden groupid (like a system account)\n");
            break;
        case  GLEXEC_MEM_ALLOC_FAILURE            :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Memory allocation failure\n");
            break;
        case  GLEXEC_INSUFFICIENT_PRIVILEGES      :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Insufficient privileges (no setuid bit on root for setuid-enabled gLExec?)\n");
            break;
        case  GLEXEC_LOG_FILE_OPEN_FAIL           :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Error: Could not open log file for gLExec.\n");
            break;
        case  GLEXEC_LOG_FILE_WRONG_PERMISSION    :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Permissions or ownership of the log file or its path were found to be wrong.\n");
            break;
        case  GLEXEC_LOG_FILE_CLOSE_FAIL          :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Can't close log file for gLExec?\n");
            break;
        case  GLEXEC_ENV_CLEAN_ERROR              :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Error during cleaning of environment.\n");
            break;
        case  GLEXEC_TARGET_ENV_ERROR              :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Error during creation of target environment.\n");
            break;
        case  GLEXEC_UNCLASIFIED                  :
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Unclassified error found. Please file a bug.\n");
            break;

        /* Authorisation failure */
        case  GLEXEC_LCAS_FAILED                  :
            /* Not all errors are LCAS / LCMAPS errors, only these are */
            snprintf (error_msg_buf, ERR_MAX_BUF, "Error in LCAS/LCMAPS, rc = %d\n", rc);
            if (glexec_get_log_to_file()) 
            {
		envval1=getenv("LCAS_LOG_FILE"); /* Note glexec_lcas() sets it */
                snprintf (error_msg_buf, ERR_MAX_BUF, "  LCAS failed, see '%s' for more info.\n",
			(envval1 ? envval1 : "(undefined)"));
            }
            else 
            {
                snprintf (error_msg_buf, ERR_MAX_BUF, "  LCAS failed.\n");
            }
            break;
        case  GLEXEC_LCMAPS_FAILED                :
            snprintf (error_msg_buf, ERR_MAX_BUF, "Error in LCAS/LCMAPS, rc = %d\n", rc);

            if (glexec_get_log_to_file()) 
            {
		envval1=getenv("LCMAPS_LOG_FILE"); /* Note glexec_lcmaps() sets it */
                snprintf (error_msg_buf, ERR_MAX_BUF, "  LCMAPS failed, see '%s' for more info.\n",
			(envval1 ? envval1 : "(undefined)"));
            }
            else 
            {
                snprintf (error_msg_buf, ERR_MAX_BUF, "  LCMAPS failed.\n");
            }
            break;
        case  GLEXEC_USER_NOT_WHITELISTED         :
	    uid=getuid();
	    pw=getpwuid(uid);
            snprintf (error_msg_buf, ERR_MAX_BUF, "  The user, uid=%d(%s), is not whitelisted. The calling user account is not privileged to execute and use gLExec.\n",
		     uid,(pw!=NULL && pw->pw_name!=NULL) ? pw->pw_name : "???");
            break;

        default:
            snprintf (error_msg_buf, ERR_MAX_BUF, "  Unknown error (%d) encountered!\n", rc);

    }

    /*
     *Everything went to the GLEXEC_LOG_ERROR anyway, 
     * now I can state more easily to which stderr or the logfile/syslog
     * And... the output is buffered and limited by the buffersize
     */
    switch (forwho)
    {
        case LOG_TO_ADMIN :
            glexec_log(GLEXEC_LOG_ERROR, "%s", error_msg_buf);
            break;
        case LOG_TO_USER :
            fprintf (stderr, "[gLExec]: %s", error_msg_buf);
            break;
        case LOG_TO_ADMIN_AND_USER :
            fprintf   (stderr, "[gLExec]: %s", error_msg_buf);
            glexec_log(GLEXEC_LOG_ERROR, "%s", error_msg_buf);
            break;
        default :
            glexec_log(GLEXEC_LOG_ERROR, "%s", error_msg_buf);
    }
}

/*!
 *  Display and log a multiline error message.
 *  \param print_err_fmt NULL terminated list of error lines.
 */
void display_and_log_error_array(char *print_err_fmt[])   {
    char *format="[gLExec]:  %s";
    int i=0;

    /* Check there is at least one message */
    if (print_err_fmt[0]==NULL)
	return;

    /* Log and print the different messages */
    do {
	glexec_log(GLEXEC_LOG_ERROR, print_err_fmt[i]);
	fprintf (stderr, format, print_err_fmt[i]);
	/* After the first entry, don't print the '[gLExec]' prefix */
	if (i==0)
	    format="         %s";
    } while (print_err_fmt[++i]!=NULL);
}

