/**
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 *  \file glexec_exit.h
 *
 *  Provide functionality for (premature) exit.
 */

#ifndef _GLEXEC_EXIT_H
#define _GLEXEC_EXIT_H

/************************************************************************/
/* DEFINES AND TYPES							*/
/************************************************************************/

typedef enum GLEXEC_MAPPED_EXIT_CODES_E
{
    GLEXEC_EXIT_USER_ERROR                  = 201,
    GLEXEC_EXIT_INTERNAL_ERROR              = 202,
    GLEXEC_EXIT_AUTHZ_ERROR                 = 203,
    GLEXEC_EXIT_CHILD_PROCESS_RC_OVERLAP    = 204,
    GLEXEC_EXIT_CHILD_EXEC_ERROR            = 205
} GLEXEC_MAPPED_EXIT_CODES_T;

/*!
 * Define of a special value that doesn't overlap with the shell's error codes
 * (0-255). */
#define GLEXEC_NO_CHILD_EXIT_CODE  257

/*!
 * Tested with bash: when an error is triggered, the execve() passed program
 * returns the exit code of '126' when the command can't be executed.
 * See also http://tldp.org/LDP/abs/html/exitcodes.html
 */
#define GLEXEC_EXECVE_FAILED_EXIT_CODE	126

/*!
 * When child exits via a signal, return GLEXEC_EXECVE_SIGFAILED_BASE+signal.
 * See also http://tldp.org/LDP/abs/html/exitcodes.html
 */
#define GLEXEC_EXECVE_SIGEXIT_BASE	128

/*!
 *  Glexec error codes.
 *  Please never change any number unless it is deemed absolutely necessary.
 *  One should always append any new error code!
 */
typedef enum {
  GLEXEC_SUCCESS                       = 100,      /*!<  General success.                                                   */
  GLEXEC_NO_SOURCE_PROXY               = 101,      /*!<  Source Proxy was not found.                                        */
  GLEXEC_STAT_ERROR                    = 102,      /*!<  Stat system call returned with an error.                           */
  GLEXEC_WRONG_PERMISSION              = 103,      /*!<  Wrong permissions bit were found.                                  */
  GLEXEC_COPY_PROXY_FAILED             = 104,      /*!<  Failed copying the proxy.                                          */
  GLEXEC_RSL_FAILED                    = 105,      /*!<  Failed to create rsl string.                                       */
  GLEXEC_LCMAPS_FAILED                 = 106,      /*!<  LCMAPS did not authorize the user.                                 */
  GLEXEC_LCAS_FAILED                   = 107,      /*!<  LCAS did not authorize the user.                                   */
  GLEXEC_OUT_OF_MEMORY                 = 108,      /*!<  Not enough memory.                                                 */
  GLEXEC_COULD_NOT_DROP_PRIVILEGES     = 109,      /*!<  Privileges could not be dropped.                                   */
  GLEXEC_PROXY_IS_LINK                 = 110,      /*!<  The source or glexec client proxy is a symbolic link and considered dangerous.      */
  GLEXEC_SOURCE_PROXY_OPEN_FAILED      = 112,      /*!<  Opening the source proxy failed.                                   */
  GLEXEC_WRONG_NUMBER_OF_BYTES_READ    = 114,      /*!>  Read wrong number of bytes.                                        */
  GLEXEC_FAILURE_WRITING_PROXY         = 115,      /*!<  A failure during writing has happened.                              */
  GLEXEC_WRONG_OWNERSHIP               = 116,      /*!<  The ownership is wrong. Applies for ownership and group.           */
  GLEXEC_PROXY_TOO_LARGE               = 117,      /*!<  The source or client proxy was found to be too large for a proxy.  */
  GLEXEC_PRIVILEGED_USER               = 118,      /*!<  Current identity of the user is that of a privileged user.         */
  GLEXEC_REJECTED_ENV_VAR              = 119,      /*!<  The environment variable contained unwanted characters.            */
  GLEXEC_FILE_CHANGED                  = 120,      /*!<  The file being used has changed for unknown reasons.               */
  GLEXEC_UNKNOWN_USER                  = 121,      /*!<  The user is unknown.                                                */
  GLEXEC_UNKNOWN_GROUP                 = 122,      /*!<  The group is unknown.                                               */
  GLEXEC_FAILED_TO_RESTORE_EFF_UID     = 123,      /*!<  Failed to set the effective uid to the original value.             */
  GLEXEC_USER_NOT_WHITELISTED          = 124,      /*!<  The user calling glexec is not in the execution whitelist.         */
  GLEXEC_MAPPED_TO_FORBIDDEN_USER      = 125,      /*!<  The mapping has resulted into a mapping to root or forbidden user  */
  GLEXEC_MAPPED_TO_FORBIDDEN_GROUP     = 126,      /*!<  The mapping has resulted into a mapping to root or forbidden group */
  GLEXEC_INSUFFICIENT_PRIVILEGES       = 127,      /*!<  The process has insufficient privileges to perform the task        */
  GLEXEC_MEM_ALLOC_FAILURE             = 128,      /*!<  Memory allocation failure, likely: insufficient free memory        */
  GLEXEC_FAILURE_READING_PROXY_ENV     = 129,      /*!<  Failure in reading the CLIENT proxy environment information        */
  GLEXEC_FAILURE_READING_PROXY_FILE    = 130,      /*!<  Failure in reading the CLIENT proxy from the file                  */
  GLEXEC_STAT_ERROR_EXECUTABLE_CMD     = 134,      /*!<  Error for the Executable command				    */
  GLEXEC_STAT_ERROR_CURRENT_WORK_DIR   = 135,      /*!<  Stat system call error for the current working directory           */
  GLEXEC_CONF_VALUE_MALFORMED          = 136,      /*!<  Configuration file has malformed values that can't be understood    */
  GLEXEC_LOG_FILE_OPEN_FAIL            = 137,      /*!<  Failed to open the glexec log file                                 */
  GLEXEC_LOG_FILE_WRONG_PERMISSION     = 138,      /*!<  Wrong permissions bit were found on the gLExec log file            */
  GLEXEC_WRONG_SET_OF_ARGS             = 140,      /*!<  Wrong set of arguments on the command line                         */
  GLEXEC_CONF_FAIL_USER_ID_SWITCH_BY   = 141,      /*!<  Failure found in the user_identity_switch_by value in conf file    */
  GLEXEC_INCOMPLETE_MAPPING_RESULT     = 142,      /*!<  Incomplete mapping result returned (could be missing user or group)*/
  GLEXEC_BACKLOG_FAILED                = 143,      /*!<  Failed to create and perform the backlog                           */
  GLEXEC_LCAS_INIT_FAIL                = 145,      /*!<  LCAS initialization failed                                         */
  GLEXEC_LCMAPS_INIT_FAIL              = 146,      /*!<  LCMAPS initialization failed                                       */
  GLEXEC_STAT_ERROR_CMD_SYMLINKED      = 147,      /*!<  Program to execute is symlinked                                    */
  GLEXEC_STAT_CMD_PERMS_CANT_EXEC      = 148,      /*!<  Program to execute is not exec-able or (mapped) user lacks sufficient permissions to execute it */
  GLEXEC_CONF_FAIL_LOCK_MECH           = 149,      /*!<  Failure found in the use_lock_mech value in conf file              */
  GLEXEC_WRONG_PERMISSION_PROXY        = 150,      /*!<  The Source or client proxy has the wrong permissions               */
  GLEXEC_WRONG_PERMISSION_ON_CMD       = 151,      /*!<  Wrong permissions on the command to execute                        */
  GLEXEC_PROXY_OPEN_FAILED             = 152,      /*!<  Opening the proxy failed.                                          */
  GLEXEC_GLEXEC_PROXY_OPEN_FAILED      = 154,      /*!<  Opening the GLEXEC client proxy failed                             */
  GLEXEC_LOCK_FAILURE                  = 155,      /*!<  Error during locking of any of the proxy files                     */
  GLEXEC_PROXY_CLOSE_FAILED            = 156,      /*!<  Closing one of the proxies failed.                                 */
  GLEXEC_LOG_FILE_CLOSE_FAIL           = 157,      /*!<  Closing the log file failed.                                       */
  GLEXEC_ENV_CLEAN_ERROR               = 158,      /*!<  Problem during backing up or cleaning environment                  */
  GLEXEC_TARGET_ENV_ERROR              = 159,      /*!<  Problem during creation of target environment                      */
  GLEXEC_SETENV_ERROR                  = 160,      /*!<  Problem during setting of environment variable                     */
  GLEXEC_USER_TRIED_V		       = 161,	   /*!<  Normal user tried the -V option				    */
  GLEXEC_CHPWD_ERROR                   = 162,      /*!<  Chpwd system call error for the target directory                   */
  GLEXEC_DIR_WRONG_OWNERSHIP           = 163,      /*!<  The ownership of the dir is wrong.				    */
  GLEXEC_PROG_WRONG_OWNERSHIP          = 164,      /*!<  The ownership of the prog is wrong.				    */
  GLEXEC_TOO_MANY_GROUPS	       = 165,	   /*!<  Number of target groups is too large.				    */
  GLEXEC_DID_NOT_USERSWITCH	       = 166,	   /*!<  No userswitch took place.					    */
  GLEXEC_WRONG_USERSWITCH	       = 167,	   /*!<  Wrong userswitch took place.					    */
  GLEXEC_DOUBLE_USERSWITCH	       = 168,	   /*!<  Userswitch happened during auth and is configured for gLExec.	    */
  GLEXEC_STILL_EFFECTIVE_ROOT	       = 169,	   /*!<  Userswitch happened but did not clear effective root.              */
  GLEXEC_ERROR_CHDIR		       = 170,	   /*!<  Chdir failed.							    */
  GLEXEC_ALLOC_ERROR                   = 171,      /*!<  Non-out-of-memory memory allocation error.                         */
  GLEXEC_UNCLASIFIED                   = -1        /*!<  Unclassified error.                                                */
} GLEXEC_ERROR;

/*************************************************************************/
/* FUNCTION DECLARATIONS                                                 */
/*************************************************************************/

/*
 * glexec_exit is called upon a (premature) exit, printing the correct messages,
 * closing the logs, tries to tidy up memory and the exits with the correct exit
 * value.
 *  \param glexec_internal_rc	Error code leading to the exit.
 *  \param child_return_code	Error code of the child process
 */
void glexec_exit(int glexec_internal_rc, int child_return_code);

/*!
 * Called when the execve fails: prints proper error message to stdout and
 * glexec log and then closes the log and exits.
 * See also http://tldp.org/LDP/abs/html/exitcodes.html
 * \param cmd command which was tried.
 * \param cwd current working directory, ignored when cmd is absolute.
 */
void glexec_execve_exit(char *cwd, char *fullcmd, char* realcmd);

#endif
