/**
 * Copyright (c) Members of the EGEE Collaboration. 2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 * Authors: Oscar Koeroo, Mischa Sall\'e
 *      NIKHEF Amsterdam, the Netherlands
 *      <grid-mw-security@nikhef.nl>  
 */

/*
 * glexec_lcmaps provides the functionality to dynamically open lcmaps and its
 * API
 */

#include "glexec_lcmaps.h"

#include "glexec.h"
#include "glexec_conf.h"
#include "glexec_log.h"
#include "glexec_environ.h"
#include "glexec_util.h"
#include "glexec_account.h"

#include "fileutil/fileutil.h"

#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>

#include <lcmaps/lcmaps_account.h>

/*************************************************************************/
/* TYPEDEFS			                                         */
/*************************************************************************/

/* Struct containing all the LCMAPS functions, will be filled using dlsym() */
typedef struct {
    void *handle,*helper;   /* handles to the LCMAPS library and its optional
			       helper (the latter is needed for the 'old' API to
			       provide missing symbols) */
    int new;		    /* whether we support the 'new' API, incl version */
    int voms_attr_check;    /* whether we support the OSG API */
    int (*lcmaps_init)(FILE*);
    int (*lcmaps_account_info_init)(lcmaps_account_info_t*);
    int (*lcmaps_return_account_from_pem)(char*,int,lcmaps_account_info_t*);
    int (*lcmaps_term)(void);
    int (*lcmaps_account_info_clean)(lcmaps_account_info_t*);
    /* Remainder is for 'new' API */
    int (*lcmaps_get_major_version)(void);
    int (*lcmaps_get_minor_version)(void);
    int (*lcmaps_get_patch_version)(void);
    /* Next three is the 'OSG API' */
    void (*lcmaps_disable_voms_attributes_verification)(void);
    void (*lcmaps_enable_voms_attributes_verification)(void);
    int (*lcmaps_is_set_to_verify_voms_attributes)(void);
} lcmaps_library_t;

/*************************************************************************/
/* INTERNAL FUNCTION DECLARATIONS                                        */
/*************************************************************************/

/*
 * Loads the LCMAPS library and its functions using dlopen() and dlsym().
 * For older versions of the LCMAPS library, it is necessary to provide the
 * functions getMajorVersion() getMinorVersion and getPatchVersion(), which can
 * be provided by LCAS. This function automatically loads LCAS when needed.
 * returns 0 on success and -1 on error.
 */
int lcmaps_load(const char *libdir, lcmaps_library_t *lcmaps);

/*
 * Closes the LCMAPS library.
 * returns 0 on success and -1 on error.
 */
int lcmaps_close(lcmaps_library_t *lcmaps);

/*
 * Setup LCMAPS environment variables
 * sets libdir (might be "") where to find lcmaps*.so
 */
void lcmaps_env_setup(char *logstring, const char **libdir);

/*
 * Open LCMAPS logging.
 * returns pointer to glexec_lcmaps_log file (might be NULL)
 */
FILE *lcmaps_log_setup(void);

/*
 * Converts an lcmaps_account_info_t into a glexec_account_t for the
 * target user.
 * returns GLEXEC_SUCCESS on success
 */
int lcmaps_get_account_info(lcmaps_account_info_t *account_info,
			    glexec_account_t *target);

/*************************************************************************/
/* FUNCTIONS			                                         */
/*************************************************************************/

/*
 * Setup, run and close LCMAPS, returning a target account
 * returns GLEXEC_SUCCESS when successful
 */
int glexec_lcmaps(char *pem_string, char *logstring,
		  glexec_account_t *target_account)	{
    const char *lcmaps_libdir,*voms;
    FILE *lcmaps_log;
    lcmaps_library_t lcmaps;
    lcmaps_account_info_t paccount_info;
    int rc;

    /* Setup LCMAPS environment */
    lcmaps_env_setup(logstring,&lcmaps_libdir);

    /* Load LCMAPS library */
    if (lcmaps_load(lcmaps_libdir,&lcmaps))
	return GLEXEC_LCMAPS_INIT_FAIL;

    /* Setup LCMAPS log */
    lcmaps_log=lcmaps_log_setup();

    /* Determine if we want to skip voms attr checking (only new API) */
    if (lcmaps.new) {
	voms=glexec_conf_value_default("glexec:lcmaps_voms_verification","yes");
	if ( (strcasecmp(voms, "no")==0) || (strcasecmp(voms, "false")==0) )	{
	    glexec_log(GLEXEC_LOG_NOTICE,
		       "Disabling LCMAPS VOMS attribute verification.\n");
	    lcmaps.lcmaps_disable_voms_attributes_verification();
	}
    }

    /* Init will return 0 on success or otherwise on failure, note that
     * glexec_lcmaps_log can be NULL in which case syslog is used.  */
    if (lcmaps.lcmaps_init(lcmaps_log))	{
	rc=GLEXEC_LCMAPS_INIT_FAIL;
	goto lcmaps_finalize_noterm;
    }

    glexec_log(GLEXEC_LOG_DEBUG, "Initialisation of LCMAPS has finished.\n");

    /* LCMAPS mapping call */
    if (lcmaps.lcmaps_account_info_init(&paccount_info) != 0) {
	glexec_log(GLEXEC_LOG_ERROR, "Error initialising LCMAPS account\n");
	rc=GLEXEC_LCMAPS_INIT_FAIL;
	goto lcmaps_finalize;
    }

    /* -1 is the mapcounter */
    if (lcmaps.lcmaps_return_account_from_pem(pem_string,-1,&paccount_info)!=0) 
    {
	lcmaps.lcmaps_account_info_clean(&paccount_info);
	rc=GLEXEC_LCMAPS_FAILED;
	goto lcmaps_finalize;
    }
    /* Mapping is successful: fill the target_account struct */
    rc=lcmaps_get_account_info(&paccount_info,target_account);

    /* Clean up lcmaps */
    lcmaps.lcmaps_account_info_clean(&paccount_info);

lcmaps_finalize:
    /* terminate LCMAPS */
    lcmaps.lcmaps_term();

lcmaps_finalize_noterm:
    /* Close lcmaps library */
    lcmaps_close(&lcmaps);

    /* Close the log file only if one is actually open (like in syslog usage
     * mode) */
    if (lcmaps_log) {
	if (fclose(lcmaps_log))
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Error while closing lcmaps log file: %s\n",
		    strerror(errno));
	lcmaps_log = NULL;
    }

    /* Now check whether LCMAPS succeeded */
    if (rc==GLEXEC_SUCCESS)
	glexec_log(GLEXEC_LOG_DEBUG, "LCMAPS is finished\n");

    return rc;
}

/*
 * Print the known LCMAPS defines to file out.
 */
void glexec_lcmaps_defines(FILE *out)	{
#ifdef LCMAPS_DB_FILE
    fprintf(out," -D LCMAPS_DB_FILE=\"%s\"\n", LCMAPS_DB_FILE);
#endif
#ifdef LCMAPS_LOG_FILE
    fprintf(out," -D LCMAPS_LOG_FILE=\"%s\"\n", LCMAPS_LOG_FILE);
#endif
#ifdef LCMAPS_DEBUG_LEVEL
    fprintf(out," -D LCMAPS_DEBUG_LEVEL=\"%s\"\n", LCMAPS_DEBUG_LEVEL);
#endif
#ifdef LCMAPS_LIBDIR
    fprintf(out," -D LCMAPS_LIBDIR=\"%s\"\n", LCMAPS_LIBDIR);
#endif
#ifdef LCMAPS_POLICY_NAME
    fprintf(out," -D LCMAPS_POLICY_NAME=\"%s\"\n", LCMAPS_POLICY_NAME);
#endif
}

/*************************************************************************/
/* INTERNAL FUNCTIONS			                                 */
/*************************************************************************/

/*
 * Loads the LCMAPS library and its functions using dlopen() and dlsym().
 * For older versions of the LCMAPS library, it is necessary to provide the
 * functions getMajorVersion() getMinorVersion and getPatchVersion(), which can
 * be provided by LCAS. This function automatically loads LCAS when needed.
 * returns 0 on success and -1 on error.
 */
int lcmaps_load(const char *libdir, lcmaps_library_t *lcmaps)   {
    char *error=dlerror(); /* Make sure dlerror() is cleared */
    const char *lcmaps_so="liblcmaps_return_account_from_pem" LIBSUFF;
    const char *helper_so="liblcas" LIBSUFF;
    char lcmaps_lib_path[AP_MAXPATH],helper_lib_path[AP_MAXPATH];
    const char *lcmaps_lib,*helper_lib;

    /* prefix lcmaps library name with libdir when needed */
    if (strlen(libdir)>0)   {
	if (snprintf(lcmaps_lib_path,AP_MAXPATH,"%s/%s",libdir,lcmaps_so)
		>=AP_MAXPATH)    {
	    glexec_log(GLEXEC_LOG_ERROR,"Full path %s/%s is too long\n",
		    libdir,lcmaps_so);
	    return -1;
	}
	lcmaps_lib=lcmaps_lib_path;
    } else
	lcmaps_lib=lcmaps_so;

    /* Make sure lcmaps->handle has welldefined value */
    lcmaps->handle=lcmaps->helper=NULL;

    /* Load the library lazily: this allows checking for missing symbols */
    glexec_log(GLEXEC_LOG_DEBUG, "Opening LCMAPS lib: %s\n",lcmaps_lib);
    if ( (lcmaps->handle=dlopen(lcmaps_lib,
			     RTLD_LAZY | RTLD_GLOBAL))==NULL )	{
	glexec_log(GLEXEC_LOG_ERROR,"Error opening LCMAPS lib: %s\n",dlerror());
	goto lcmaps_dlopen_failed;
    }

    /* See if we have lcmaps_get*version() functions */
    lcmaps->lcmaps_get_major_version=
	(int(*)(void))dlsym(lcmaps->handle,"lcmaps_get_major_version");
    if (dlerror()==NULL)    { /* New LCMAPS version */
	lcmaps->new=1;
	glexec_log(GLEXEC_LOG_DEBUG,"Found new LCMAPS version.\n");
    } else {	/* Old LCMAPS version */
	lcmaps->new=0;
	glexec_log(GLEXEC_LOG_DEBUG,
	    "Old LCMAPS version, "\
	    "loading LCAS to provide missing get*VersionNumber() functions.\n");
	/* prefix helper library name with libdir when needed */
	if (strlen(libdir)>0)   {
	    if (snprintf(helper_lib_path,AP_MAXPATH,"%s/%s",libdir,helper_so)
		    >=AP_MAXPATH)    {
		glexec_log(GLEXEC_LOG_ERROR,"Full path %s/%s is too long\n",
			libdir,helper_so);
		return -1;
	    }
	    helper_lib=helper_lib_path;
	} else
	    helper_lib=helper_so;
	/* now open the helper lib */
	if ( (lcmaps->helper=dlopen(helper_lib, RTLD_NOW | RTLD_GLOBAL))==NULL )    {
	    glexec_log(GLEXEC_LOG_WARNING,"Error opening %s lib: %s\n",
			helper_lib,dlerror());
	    goto lcmaps_dlopen_failed;
	}
    }

    /* Reopening lcmaps with RTLD_NOW to force loading of secondary dependencies
     * (e.g. VOMSAPI) */
    glexec_log(GLEXEC_LOG_DEBUG,
	    "Reopening LCMAPS lib %s to resolve secondary dependencies.\n",
	    lcmaps_lib);
    if (dlclose(lcmaps->handle))    {
	glexec_log(GLEXEC_LOG_ERROR,"Error closing LCMAPS lib: %s\n",dlerror());
	lcmaps->handle=NULL; /* Put to zero by hand, it might be corrupt */
	goto lcmaps_dlopen_failed;
    }
    if ( (lcmaps->handle=dlopen(lcmaps_lib,
			     RTLD_NOW | RTLD_GLOBAL))==NULL )	{
	glexec_log(GLEXEC_LOG_ERROR,"Error opening LCMAPS lib: %s\n",dlerror());
	goto lcmaps_dlopen_failed;
    }

    /* Load all functions */

    /* Standard functions */
    glexec_log(GLEXEC_LOG_DEBUG, "Loading LCMAPS symbols.\n");
    lcmaps->lcmaps_init=(int (*)(FILE*)) dlsym(lcmaps->handle,"lcmaps_init");
    if ((error=dlerror())!=NULL) goto lcmaps_dlsym_failed;
    lcmaps->lcmaps_account_info_init=(int (*)(lcmaps_account_info_t*))
	    dlsym(lcmaps->handle,"lcmaps_account_info_init");
    if ((error=dlerror())!=NULL) goto lcmaps_dlsym_failed;
    lcmaps->lcmaps_return_account_from_pem=
	(int (*)(char*,int,lcmaps_account_info_t*))
	    dlsym(lcmaps->handle,"lcmaps_return_account_from_pem");
    if ((error=dlerror())!=NULL) goto lcmaps_dlsym_failed;
    lcmaps->lcmaps_term=(int (*)(void))dlsym(lcmaps->handle,"lcmaps_term");
    if ((error=dlerror())!=NULL) goto lcmaps_dlsym_failed;
    lcmaps->lcmaps_account_info_clean=(int (*)(lcmaps_account_info_t*))
	    dlsym(lcmaps->handle,"lcmaps_account_info_clean");

    /* New API functions */
    if (lcmaps->new)	{
	lcmaps->lcmaps_get_minor_version=(int(*)(void))
	    dlsym(lcmaps->handle,"lcmaps_get_minor_version");
	if ((error=dlerror())!=NULL) goto lcmaps_dlsym_failed;
	lcmaps->lcmaps_get_patch_version=(int(*)(void))
	    dlsym(lcmaps->handle,"lcmaps_get_patch_version");
	if ((error=dlerror())!=NULL) goto lcmaps_dlsym_failed;

	/* Now check for the 'OSG functionality' */
	lcmaps->lcmaps_disable_voms_attributes_verification=(void(*)(void))
	    dlsym(lcmaps->handle,"lcmaps_disable_voms_attributes_verification");
	if (dlerror()==NULL) { /* OSG functionality is present */
	    lcmaps->voms_attr_check=1;
	    glexec_log(GLEXEC_LOG_DEBUG,
		"LCMAPS API for dis-/enabling voms attr checking found.\n");
	    /* Now open the rest, errors are real... */
	    lcmaps->lcmaps_enable_voms_attributes_verification=(void(*)(void))
		dlsym(lcmaps->handle,
		      "lcmaps_enable_voms_attributes_verification");
	    if ((error=dlerror())!=NULL) goto lcmaps_dlsym_failed;
	    lcmaps->lcmaps_is_set_to_verify_voms_attributes=(int(*)(void))
		dlsym(lcmaps->handle,"lcmaps_is_set_to_verify_voms_attributes");
	    if ((error=dlerror())!=NULL) goto lcmaps_dlsym_failed;
	} else	{ /* OSG functionality is absent: not an error */
	    lcmaps->voms_attr_check=0;
	    glexec_log(GLEXEC_LOG_DEBUG,
		"LCMAPS API for dis-/enabling voms attr checking NOT found.\n");
	}
    }
    return 0;

lcmaps_dlsym_failed:
    glexec_log(GLEXEC_LOG_ERROR,"Error loading LCMAPS function: %s\n",error);

lcmaps_dlopen_failed:
    if (lcmaps->handle)	dlclose(lcmaps->handle);
    if (lcmaps->helper) dlclose(lcmaps->helper);
    lcmaps->helper=lcmaps->handle=NULL;
    return -1;
}

/*
 * Closes the LCMAPS library.
 * returns 0 on success and -1 on error.
 */
int lcmaps_close(lcmaps_library_t *lcmaps)	{
    int rc1=0,rc2=0;

    glexec_log(GLEXEC_LOG_DEBUG,"Closing LCMAPS library.\n");
    if ((rc1=dlclose(lcmaps->handle)))
	glexec_log(GLEXEC_LOG_ERROR,"Error closing LCMAPS library: %s\n",
		    dlerror());
    lcmaps->handle=NULL;
    if (lcmaps->helper)	{
	glexec_log(GLEXEC_LOG_DEBUG,"Closing LCMAPS helper library.\n");
	if ((rc2=dlclose(lcmaps->helper)))
	    glexec_log(GLEXEC_LOG_ERROR,
			"Error closing LCMAPS helper library: %s\n",
			dlerror());
	lcmaps->helper=NULL;
    }
    return (rc1 || rc2 ? -1 : 0);
}

/*
 * Setup LCMAPS environment variables
 * set libdir (might be "") where to find lcmaps*.so
 */
void lcmaps_env_setup(char *logstring, const char **libdir) {
    const char *moddir_suffix="/modules";
    char modules_dir[AP_MAXPATH];
    struct stat st;

    /* Get LCMAPS variables */
    setenv("LCMAPS_LOG_STRING", logstring , 1);

    /* setenv("LCMAPS_VERIFY_TYPE", "uid"); */
    /* Disabling the blocking factor to use multiple LCMAPS policies */
    setenv("LCMAPS_VERIFY_TYPE", "uid_pgid", 1);

    glexec_setenv_regex("LCMAPS_DEBUG_LEVEL",
	glexec_conf_value_default("glexec:lcmaps_debug_level",LCMAPS_DEBUG_LEVEL),
	"^[0-5]$");
    glexec_setenv_regex("LCMAPS_DB_FILE",
	glexec_conf_value_default("glexec:lcmaps_db_file",LCMAPS_DB_FILE),
	"\\/(.*\\/)+[^/]*");

    /* Setup LCMAPS policy */
    setenv( "LCMAPS_POLICY_NAME",
	    glexec_conf_value_default("glexec:lcmaps_get_account_policy",
					LCMAPS_POLICY_NAME),
	    1 );

    /* Setup LCMAPS libdir, which will be used for dlopen-ing LCMAPS */
    *libdir=glexec_conf_value_default("glexec:lcmaps_libdir",LCMAPS_LIBDIR);

    /* If it's non-empty, check it is an absolute and valid directory */
    if ( (*libdir)[0]!='\0')   {
	if ( (*libdir)[0]!='/' || stat(*libdir,&st)!=0 || !S_ISDIR(st.st_mode)) {
	    glexec_log(GLEXEC_LOG_WARNING,
		"Warning: ignoring lcmaps_libdir config option: "\
		"%s is not an absolute path to a valid directory.\n", *libdir);
	    *libdir=LCMAPS_LIBDIR;
	}
    }
    setenv("LCMAPS_LIBDIR", *libdir, 1);

    /* If it's non-empty, also setup LCMAPS_MODULES_DIR. */
    if ( (*libdir)[0]!='\0')    { /* No need for strlen... */
	if (snprintf(modules_dir,AP_MAXPATH,"%s%s",
		     *libdir, moddir_suffix) >= AP_MAXPATH)
	    glexec_log(GLEXEC_LOG_WARNING,
		"Full modulespath '%s%s' would be too long, " \
		"not setting LCMAPS_MODULES_DIR\n",
		libdir,moddir_suffix);
	else	{
	    glexec_log(GLEXEC_LOG_DEBUG,"Setting LCMAPS_MODULES_DIR to '%s'\n",
		       modules_dir);
	    setenv("LCMAPS_MODULES_DIR", modules_dir, 1);
	}
    }

}

/*
 * Open LCMAPS logging.
 * returns pointer to glexec_lcmaps_log file (might be NULL)
 */
FILE *lcmaps_log_setup(void) {
    const mode_t file_mask=GLEXEC_LOG_FILEMASK; /* File permissions */
    const mode_t dir_mask=GLEXEC_LOG_DIRMASK; /* Dir permissions */
    int rc;
    gid_t oldgid=getegid(),lcmaps_loggid;
    char *lcmaps_log_name;
    FILE *lcmaps_log=NULL;

    /* Optionally open logfile */
    if (glexec_get_log_to_syslog()) { /* Log to SysLog */
	/*  Make sure the environment variables are never set when glexec has to
	 *  use syslog.*/
        unsetenv("LCMAPS_LOG_FILE");
    } else { /* logfile */
	glexec_setenv_regex("LCMAPS_LOG_FILE",
	    glexec_conf_value_default("glexec:lcmaps_log_file",LCMAPS_LOG_FILE),
	    "\\/(.*\\/)+[^/]*");
        if ( (lcmaps_log_name = getenv("LCMAPS_LOG_FILE"))==NULL )   {
            glexec_log(GLEXEC_LOG_ERROR,
		"Log file for LCMAPS could not be opened: "\
		"No log file name was specified.\n");
            glexec_exit(GLEXEC_LCMAPS_INIT_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	}
	/* Get logging GID as set by glexec */
	lcmaps_loggid=glexec_get_loggid();
	/* Set the effective gid, so that file and dir will be created with the
	 * right ownership */
	if (setegid(lcmaps_loggid))
	    glexec_log(GLEXEC_LOG_WARNING,
		"Setting effective gid to %d failed.\n", (int)lcmaps_loggid);
	/* Now try to open the logfile with file permission file_mask and directory
	 * permissions dir_mask. Don't force it to be trusted. */
	rc=cgul_open_logfile(lcmaps_log_name, file_mask, dir_mask, &lcmaps_log, 0);
	/* reset effective gid */
	if (setegid(oldgid))
	    glexec_log(GLEXEC_LOG_WARNING,
		"Setting effective gid to %d failed.\n", (int)oldgid);
	/* Parse return code of cgul_open_logfile */
	switch (rc)	{
	    case 0: break;
	    case -1: 
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: error creating dir for LCMAPS logfile '%s'.\n",
		    lcmaps_log_name);
		glexec_exit(GLEXEC_LCMAPS_INIT_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	    case -2:
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: I/O error during opening LCMAPS logfile: %s.\n",
		    strerror(errno));
		glexec_exit(GLEXEC_LCMAPS_INIT_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	    case -3:
		glexec_log(GLEXEC_LOG_WARNING,
		    "WARNING: Permissions for LCMAPS logfile are wrong " \
		    "(should be: user root, " \
		    "permissions: file at most 750, dir at most 755). "\
		    "In next version this will be a fatal error.\n");
		break;
	    case -4:
		glexec_exit(GLEXEC_OUT_OF_MEMORY, GLEXEC_NO_CHILD_EXIT_CODE);
	    case -5:
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: LCMAPS logfile '%s' is not an absolute path.\n",
		    lcmaps_log_name);
		glexec_exit(GLEXEC_LCMAPS_INIT_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	    case -6:
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: LCMAPS logfile is no longer the one that was opened!.\n");
		glexec_exit(GLEXEC_LCMAPS_INIT_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	    case -7:
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: Specified LCMAPS logfile is not a regular file.\n");
		glexec_exit(GLEXEC_LCMAPS_INIT_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	    case -8:
	    default:
		glexec_exit(GLEXEC_LCMAPS_INIT_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	}

	/* Log the log filename */
	glexec_log(GLEXEC_LOG_DEBUG,
	    "LCMAPS will be initialised with log file: %s\n", lcmaps_log_name);
    }

    return lcmaps_log;
}

/*
 * Converts an lcmaps_account_info_t into a glexec_account_t for the
 * target user.
 * returns GLEXEC_SUCCESS on success
 */
int lcmaps_get_account_info(lcmaps_account_info_t *account_info,
			    glexec_account_t *target) {
    struct passwd *pw;
    struct group *pgr,*sgr;
    int i;

    /* UID */
    if ( (pw = glexec_getpwuid(account_info->uid))==NULL )	{
	glexec_log(GLEXEC_LOG_ERROR,
	      "Error looking up uid (%d)\n", account_info->uid);
	return GLEXEC_UNKNOWN_USER;
    }
    target->uid = account_info->uid;
    target->uname = pw->pw_name;
    /* homedir */
    target->dir = pw->pw_dir;
    if (target->dir==NULL)
	glexec_log(GLEXEC_LOG_DEBUG,
		"User %s(%s) has no home directory.\n",
		target->uid,
		target->uname ? target->uname : "unknown");

    /* primary GID */
    if (account_info->npgid == 0) {
	glexec_log(GLEXEC_LOG_ERROR, "Need at least 1 gid!\n");
	return GLEXEC_INCOMPLETE_MAPPING_RESULT;
    }
    if ( (pgr = glexec_getgrgid(account_info->pgid_list[0]))==NULL ) {
	glexec_log(GLEXEC_LOG_ERROR,
		"Error looking up gid (%d)\n", account_info->pgid_list[0]);
	return GLEXEC_UNKNOWN_GROUP;
    }
    target->gid = account_info->pgid_list[0];
    target->gname = pgr->gr_name;

    /* Print out of uid and primary GID */
    glexec_log(GLEXEC_LOG_INFO,
		"Found uid = %d(%s), gid = %d(%s)\n",
		target->uid, target->uname, target->gid, target->gname);

    /* secondary GIDs */
    if ( (target->nsgid = account_info->nsgid) > 0) {
	glexec_log(GLEXEC_LOG_INFO, "Found secondary gids:\n");
	target->sgid = (gid_t *) gmalloc (sizeof (gid_t) * target->nsgid);

	for (i = 0; i < target->nsgid; i++) {
	    /* First copy the GID itself */
	    target->sgid[i] = account_info->sgid_list[i];
	    /* Get account information if available */
	    if ( (sgr=glexec_getgrgid(target->sgid[i]))==NULL )  {
		/* Print notice, but include gid */
		glexec_log(GLEXEC_LOG_NOTICE,
		    "Could not look up name for sgid (%d)\n",
		    target->sgid[i]);
	    } else {
		/* Print found secondary GID and GROUP */
		glexec_log(GLEXEC_LOG_INFO,
		    "           gid = %d(%s)\n",
		    target->sgid[i], sgr->gr_name);
	    }
	}
    }

    return GLEXEC_SUCCESS;
}

