/**
 * Copyright (c) Members of the EGEE Collaboration. 2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 * Authors: Oscar Koeroo, Mischa Sall\'e
 *      NIKHEF Amsterdam, the Netherlands
 *      <grid-mw-security@nikhef.nl>  
 */

/*!
 * \file glexec_log.c provides the logging functionality.
 */

#include "glexec_log.h"

#include "glexec.h"
#include "glexec_exit.h"
#include "glexec_conf.h"
#include "glexec_util.h"
#include "glexec_account.h"
#include "fileutil/fileutil.h"

#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <syslog.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <grp.h>

/*************************************************************************/
/* DEFINES								 */
/*************************************************************************/

/* Size of 33 is good enough for PID length of 5 and name glexec, just make it
 * somewhat bigger */
#define GLEXEC_LOGPREFIX_LENGTH	50

/* Prepend syslog entries with following string */
#ifndef GLEXEC_SYSLOG_IDENT
#define GLEXEC_SYSLOG_IDENT "glexec"
#endif

/* Options to use when opening syslog */
#ifndef GLEXEC_SYSLOG_OPTS
#define GLEXEC_SYSLOG_OPTS  (LOG_CONS|LOG_PID)
#endif

/* Default syslog priority  */
#ifndef GLEXEC_SYSLOG_DEF_PRIORITY
#define GLEXEC_SYSLOG_DEF_PRIORITY LOG_ERR
#endif

/*************************************************************************/
/* TYPES								 */
/*************************************************************************/

struct glexec_message {
  char *timestamp;
  char *message;
  struct glexec_message *next;
};

/*************************************************************************/
/* GLOBAL VARIABLES							 */
/*************************************************************************/

/* Default log level NOTICE */
static glexec_log_level_t glexec_log_level = GLEXEC_LOG_NOTICE;

/* Default destination is known and syslog */
static int glexec_log_to_file	= 0;  /*! Default is NOT to log to a file
					directly, \see glexec_log_to_syslog */
static int glexec_log_to_syslog = 1;  /*! Default is to log to syslog. \see
					glexec_log_to_file */
static int glexec_syslog_opened	= 0;

/* Whether glexec's internal loglevels correspond to different syslog levels.
 * Default is 0, which means all will be GLEXEC_SYSLOG_DEF_PRIORITY */
static int glexec_diff_syslog_levels = 0;

/* Which type of syslog facility to use, default is GLEXEC_SYSLOG_FACILITY */
static int glexec_syslog_facility = GLEXEC_SYSLOG_FACILITY;

/* glexec log file */
static FILE *glexec_log_file    = NULL;

/* glexec log gid: also used for LCAS & LCMAPS, can be obtained using getter
 * glexec_get_loggid() */
static gid_t glexec_loggid=0;

/* Mapping from gLExec log levels to syslog levels */
static const int glexec_syslog_level[GLEXEC_LOG_LEVEL_HIGH]={
    	/* LOG_EMERG unused */
	/* LOG_ALERT unused */ 
	LOG_CRIT,	/* GLEXEC_LOG_LEVEL_LOW */
	LOG_ERR,	/* GLEXEC_LOG_ERROR */
	LOG_WARNING,	/* GLEXEC_LOG_WARNING */
	LOG_NOTICE,	/* GLEXEC_LOG_NOTICE */
	LOG_INFO,	/* GLEXEC_LOG_INFO */
	LOG_DEBUG};	/* GLEXEC_LOG_DEBUG */

/*************************************************************************/
/* FUNCTION PROTOTYPES							 */
/*************************************************************************/

/* Open logfile */
void glexec_open_log_file(void);

/* Open syslog */
void glexec_open_syslog(void);

/*************************************************************************/
/* FUNCTIONS								 */
/*************************************************************************/

/*!
 * Set logging to syslog (unsets to file)
 */
void glexec_set_log_to_syslog(void) {
    glexec_log_to_syslog=1;
    glexec_log_to_file=0;
}

/* Get whether logging to syslog is enabled */
int glexec_get_log_to_syslog(void)  {
    return glexec_log_to_syslog;
}

/*!
 * Set logging to file (unsets to syslog)
 */
void glexec_set_log_to_file(void)   {
    glexec_log_to_syslog=0;
    glexec_log_to_file=1;
}

/*!
 * Get whether logging to file is enabled
 * */
int glexec_get_log_to_file(void)  {
    return glexec_log_to_file;
}

/*!
 * Getter method for glexec logfile gid
 * \return glexec log gid
 */
gid_t glexec_get_loggid(void)	{
    return glexec_loggid;
}

/*!
 *  Set the log level for glexec taking into account the lowest and
 *  highest possible log level.
 *  \param loglevel  New log level.
 *  \param diff_syslog_levels  when 1 glexec_log_levels will be translated into
 *         different syslog levels (only in case logging to syslog of course).
 */
void glexec_set_log_level(glexec_log_level_t loglevel, int diff_syslog_levels) {
    /* Set whether to use different syslog levels */
    glexec_diff_syslog_levels=diff_syslog_levels;

    /* truncate a too small loglevel */
    if (loglevel < GLEXEC_LOG_LEVEL_LOW)    {
	glexec_log_level = GLEXEC_LOG_LEVEL_LOW;
	glexec_log(GLEXEC_LOG_WARNING,
	    "Specified loglevel is too small, using min. log level %d.\n",
	    GLEXEC_LOG_LEVEL_LOW);
    } else {
	if (loglevel>=GLEXEC_LOG_LEVEL_HIGH) {
	    /* truncate a too large loglevel */
	    glexec_log_level = GLEXEC_LOG_LEVEL_HIGH;
	    glexec_log(GLEXEC_LOG_WARNING,
		"Specified loglevel is too large, using max. log level %d.\n",
		GLEXEC_LOG_LEVEL_HIGH);
	} else {
	    glexec_log_level = loglevel;
	    glexec_log(GLEXEC_LOG_INFO, "Setting log level to: %d\n",
		       loglevel);
	}
    }

    /* Only now log whether we differentiate, so that we take the loglevel into
     * account */
    if (diff_syslog_levels)
	glexec_log(GLEXEC_LOG_DEBUG, "Using different syslog loglevels.\n");
}

/*!
 *  Close the log file and syslog if it had been opened.
 */
void glexec_close_log(void) {
    /* syslog */
    if (glexec_syslog_opened) {
        closelog();
        glexec_syslog_opened = 0;
    }

    /* file */
    if (glexec_log_file) {
	/* if it fails, we can (hopefully) still log... */
        if (fclose(glexec_log_file))
	    glexec_log(GLEXEC_LOG_ERROR,
		"Error while closing glexec log file: %s\n", strerror(errno));
        glexec_log_file = NULL;
    } 
}

 
/*!
 * Wrapper function which will try to open the correct log destination
 */
void glexec_open_log(void)  {
    if (glexec_log_to_file) { /* We want log to file */
	/* Make sure that we change to syslog for logging-only mode as soon as
	 * possible, to prevent lcas/lcmaps opening their logs.*/
	if (!glexec_is_switching())	{
	    /* Logging only mode: CANNOT log to file */
	    glexec_open_syslog();
	    glexec_log(GLEXEC_LOG_ERROR,
			"ERROR: Cannot log to file in logging-only mode, " \
			"logging to syslog from now on.\n");
	    return;
	}
	/* Now try opening the logfile */
	glexec_open_log_file();
	/* It might fail (e.g. no longer effective root), then keep syslog */
	if (glexec_log_file==NULL)  {
	    glexec_open_syslog();
	    glexec_log(GLEXEC_LOG_ERROR,
		"ERROR: Cannot open logfile, logging to syslog from now on.\n");
	}
    } else /* Log to syslog */
	glexec_open_syslog();
}

/*!
 * Glexec log function, automatically opens the log when needed.
 * When logging to file is specified but fails, it will change to syslog.
 * \param loglevel  The log level of this message.
 * \param fmt       The log message itself
 */
void glexec_log(glexec_log_level_t loglevel, const char* fmt, ...) {
    pid_t glexec_pid;
    char prefix[GLEXEC_LOGPREFIX_LENGTH];
    char body[GLEXEC_MAX_LOG_LINE];
    int i,prefix_length=0, body_length, prefix_too_long=0, log_failed=0,
	syslog_level;

    time_t timevar;
    struct tm *lt;
    va_list ap;

    /*  Check the log level of the message. If it is greater then the allowed
     *  log level, do not display this message. Also make sure that we do not
     *  log any null messages.
     */
    if ((loglevel > glexec_log_level) || (fmt == NULL) || (*fmt == '\0'))
	return;

    /* Initialize glexec_pid: do this each time to make diff between child and
     * parent */
    glexec_pid=getpid();
    
    /* Real message: Put in buffer for later use */
    va_start(ap, fmt);
    body_length = vsnprintf(body, GLEXEC_MAX_LOG_LINE, fmt, ap);
    if (body_length>=GLEXEC_MAX_LOG_LINE) { /* Logline too long: truncate */
	sprintf(&(body[GLEXEC_MAX_LOG_LINE-5]),"...\n");
	body_length=GLEXEC_MAX_LOG_LINE-1;
    }
    va_end(ap);

    /* Replace '?' for any non-printable character */
    for (i=0; i<body_length-1; i++)   {
	if (!isprint(body[i]))
	    body[i]='?';
    }
    /* Force a newline at the end. */
    body[body_length-1]='\n';

    /* In case we are logging to a file, prefix the message by a time stamp.
     * The syslog facility does this automatically and hence we should not
     * prefix it in that case. */
    if (glexec_log_to_file) { /* Log to file */
	/* Setup prefix */
	time(&timevar);
	lt = gmtime(&timevar);
	prefix_length = snprintf(prefix, GLEXEC_LOGPREFIX_LENGTH,
		"%s[%d] %d%.2d%.2dT%.2d%.2d%.2dZ: ",
		GLEXEC_PACKAGE_NAME, (int)glexec_pid,
		lt->tm_year + 1900, lt->tm_mon + 1, lt->tm_mday,
		lt->tm_hour, lt->tm_min, lt->tm_sec);
	/* If prefix doesn't fit: switch to syslog. Print warning afterwards */
	if (prefix_length>=GLEXEC_LOGPREFIX_LENGTH)
	    prefix_too_long=1;
	else	{
	    /* Log the actual message: if it fails switch to syslog */
	    if (fprintf(glexec_log_file, "%s%s", prefix, body)<0)
		log_failed=1;
	    else    {
		/* Succesfully logged to file */
		fflush(glexec_log_file);
		return;
	    }
	}
	/* Do the switch to syslog and log the reason */
	fclose(glexec_log_file); glexec_log_file=NULL;
	glexec_open_syslog(); /* This also runs glexec_set_log_to_syslog() */
	/* Log reason for changing */
	if (prefix_too_long)
	    syslog(LOG_ERR,
		"ERROR: Logging to file failed (prefixlength=%d is too long), " \
		"logging to syslog from now on.\n", prefix_length);
	else if (log_failed)
	    syslog(LOG_ERR,
		"ERROR: Logging to file failed, logging to syslog from now on.\n");
    } else
	/* Make sure syslog is opened, we might be logging before it's opened*/
	glexec_open_syslog();

    /* Log the actual message, note: when glexec_ */
    syslog_level= glexec_diff_syslog_levels ? glexec_syslog_level[loglevel]
					    : GLEXEC_SYSLOG_DEF_PRIORITY;
    syslog(syslog_level, "%s", body);
}

/*!
 * Set the correct syslog facility using specified facility, and reopen when
 * it was already open. When facility is NULL return the current one. Default is
 * GLEXEC_SYSLOG_FACILITY.
 * \param facility *char of facility to use, when NULL don't change.
 * \return facility new integer facility or current when facility == NULL.
 */
int glexec_set_syslog_facility(const char *facility) {
    /* nothing to do */
    if (facility==NULL)
	return glexec_syslog_facility;

    /* Now go over all the possibilities */
    if (strcmp(facility,"LOG_KERN")==0)
	glexec_syslog_facility=LOG_KERN;
    else if (strcmp(facility,"LOG_USER")==0)
	glexec_syslog_facility=LOG_USER;
    else if (strcmp(facility,"LOG_MAIL")==0)
	glexec_syslog_facility=LOG_MAIL;
    else if (strcmp(facility,"LOG_NEWS")==0)
	glexec_syslog_facility=LOG_NEWS;
    else if (strcmp(facility,"LOG_UUCP")==0)
	glexec_syslog_facility=LOG_UUCP;
    else if (strcmp(facility,"LOG_DAEMON")==0)
	glexec_syslog_facility=LOG_DAEMON;
    else if (strcmp(facility,"LOG_AUTH")==0)
	glexec_syslog_facility=LOG_AUTH;
    else if (strcmp(facility,"LOG_CRON")==0)
	glexec_syslog_facility=LOG_CRON;
    else if (strcmp(facility,"LOG_LPR")==0)
	glexec_syslog_facility=LOG_LPR;
    else if (strcmp(facility,"LOG_LOCAL0")==0)
	glexec_syslog_facility=LOG_LOCAL0;
    else if (strcmp(facility,"LOG_LOCAL1")==0)
	glexec_syslog_facility=LOG_LOCAL1;
    else if (strcmp(facility,"LOG_LOCAL2")==0)
	glexec_syslog_facility=LOG_LOCAL2;
    else if (strcmp(facility,"LOG_LOCAL3")==0)
	glexec_syslog_facility=LOG_LOCAL3;
    else if (strcmp(facility,"LOG_LOCAL4")==0)
	glexec_syslog_facility=LOG_LOCAL4;
    else if (strcmp(facility,"LOG_LOCAL5")==0)
	glexec_syslog_facility=LOG_LOCAL5;
    else if (strcmp(facility,"LOG_LOCAL6")==0)
	glexec_syslog_facility=LOG_LOCAL6;
    else if (strcmp(facility,"LOG_LOCAL7")==0)
	glexec_syslog_facility=LOG_LOCAL7;
    /* Now the non-standard ones... */
#ifdef LOG_SYSLOG
    else if (strcmp(facility,"LOG_SYSLOG")==0)
	glexec_syslog_facility=LOG_SYSLOG;
#endif
#ifdef LOG_FTP
    else if (strcmp(facility,"LOG_FTP")==0)
	glexec_syslog_facility=LOG_FTP;
#endif
#ifdef LOG_AUTHPRIV
    else if (strcmp(facility,"LOG_AUTHPRIV")==0)
	glexec_syslog_facility=LOG_AUTHPRIV;
#endif
#ifdef LOG_AUDIT
    else if (strcmp(facility,"LOG_AUDIT")==0)
	glexec_syslog_facility=LOG_AUDIT;
#endif
    else
	glexec_syslog_facility=-1;

    /* Any need to change ? */
    if (glexec_syslog_facility==GLEXEC_SYSLOG_FACILITY)
	return glexec_syslog_facility;
   
    /* Is it valid? */
    if (glexec_syslog_facility==-1)  {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Unknown syslog glexec_syslog_facility '%s' specified.\n",facility);
	glexec_exit(GLEXEC_CONF_VALUE_MALFORMED, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Reopen with new glexec_syslog_facility */
    if (glexec_syslog_opened)	{
	closelog();
	openlog(GLEXEC_SYSLOG_IDENT,GLEXEC_SYSLOG_OPTS,glexec_syslog_facility);
	glexec_log(GLEXEC_LOG_WARNING,
	    "Syslog has been reopened with new facility '%s'\n",facility);
    }
    return glexec_syslog_facility;
}

/*************************************************************************/
/* INTERNAL FUNCTIONS							 */
/*************************************************************************/

/*!
 * Open the syslog facility for gLExec.
 */
void glexec_open_syslog(void)	{
    /* Open syslog in case it wasn't open */
    if (!glexec_syslog_opened)	{
	openlog(GLEXEC_SYSLOG_IDENT,GLEXEC_SYSLOG_OPTS,glexec_syslog_facility);
	glexec_syslog_opened = 1;
    }
    /* Force setting logging to syslog. */
    glexec_set_log_to_syslog();
}

/*!
 *  Open the glexec logfile. During opening logging destination is forced to
 *  syslog, in order to record errors. Upon success glexec_log_file will be
 *  non-NULL, upon failure, glexec_log_file is still NULL, and in case it is not
 *  due to no-longer-effective-root, glexec will exit. 
 */
void glexec_open_log_file(void)	{
    const mode_t file_mask=GLEXEC_LOG_FILEMASK; /* File permissions */
    const mode_t dir_mask=GLEXEC_LOG_DIRMASK; /* Dir permissions */
    const uid_t glexec_loguid=0;
    const char *log_file,*group;
    int rc;
    struct group *gr;
    uid_t oldeuid=geteuid();
    gid_t oldegid=getegid();

    /* Temporarily set logging to default syslog */
    glexec_set_log_to_syslog();

    /* Check if there is an alternative defined for the default log file
     * for glexec. */
    log_file=glexec_conf_value_default("glexec:log_file", GLEXEC_LOG_FILE);

    /* Check if we specified a group for creation dir/file, otherwise it will be
     * group root */
    if ( (group=glexec_conf_value("glexec:log_file_group")) )  {
	if ( (gr = glexec_getgrnam(group))!=NULL )	{
	    glexec_loggid=gr->gr_gid;
	    glexec_log(GLEXEC_LOG_DEBUG,
		"Logfile will be created (if needed) using gid=%d(%s).\n",
		(int)glexec_loggid,group);
	} else /* group does not exist */
	    glexec_log(GLEXEC_LOG_WARNING,
		"Specified glexec:log_file_group '%s' not found on system, " \
		"using GID %d instead\n", group, (int)glexec_loggid);
    } else  /* Default group */
	glexec_log(GLEXEC_LOG_DEBUG,
	    "Logfile will be created (if needed) using GID %d.\n",
	    (int)glexec_loggid);

    /* Temporarily set back effective root uid (if it isn't already) */
    if (oldeuid!=glexec_loguid && seteuid(glexec_loguid))   {
	glexec_log(GLEXEC_LOG_WARNING,
	    "Setting effective uid to %d failed, reverting to syslog!\n",
	    (int)glexec_loguid);
	return;
    }
    /* Set the effective gid, so that file and dir will be created with the
     * right ownership */
    if (setegid(glexec_loggid))
	glexec_log(GLEXEC_LOG_WARNING,
	    "Setting effective gid to %d failed.\n", (int)glexec_loggid);
    /* Now try to open the logfile with file permission file_mask and directory
     * permissions dir_mask. Force it to be trusted: last arg == 1 */
    rc=cgul_open_logfile(log_file, file_mask, dir_mask, &glexec_log_file, 1);
    /* reset effective gid */
    if (setegid(oldegid))
	glexec_log(GLEXEC_LOG_WARNING,
	    "Setting effective gid to %d failed.\n", (int)oldegid);
    /* reset effective uid */
    if (seteuid(oldeuid))
	glexec_log(GLEXEC_LOG_WARNING,
	    "Setting effective uid to %d failed.\n", (int)oldeuid);
    /* Parse return code of cgul_open_logfile */
    switch (rc)	{
	case 0: break;
	case -1: 
	    glexec_log(GLEXEC_LOG_ERROR,
		"Error: error creating dir for logfile '%s'.\n", log_file);
	    glexec_exit(GLEXEC_LOG_FILE_OPEN_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	case -2:
	    glexec_log(GLEXEC_LOG_ERROR,
		"Error: I/O error during opening logfile: %s.\n",
		strerror(errno));
	    glexec_exit(GLEXEC_LOG_FILE_OPEN_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	case -3:
	    glexec_exit(GLEXEC_LOG_FILE_WRONG_PERMISSION, 
		        GLEXEC_NO_CHILD_EXIT_CODE);
	case -4:
	    glexec_exit(GLEXEC_OUT_OF_MEMORY, GLEXEC_NO_CHILD_EXIT_CODE);
	case -5:
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Error: logfile '%s' is not an absolute path.\n",log_file);
	    glexec_exit(GLEXEC_LOG_FILE_OPEN_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	case -6:
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Error: logfile is no longer the one that was opened!.\n");
	    glexec_exit(GLEXEC_LOG_FILE_OPEN_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	case -7:
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Error: logfile '%s' is not a regular file.\n",log_file);
	    glexec_exit(GLEXEC_LOG_FILE_OPEN_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	case -8:
	default:
	    glexec_exit(GLEXEC_LOG_FILE_OPEN_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Reset logging to file */
    glexec_set_log_to_file();
    if (glexec_syslog_opened) /* Close syslog if it was opened */
	closelog();
}

