/**
 * Copyright (c) Members of the EGEE Collaboration. 2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 * Authors: Oscar Koeroo, Mischa Sall\'e
 *      NIKHEF Amsterdam, the Netherlands
 *      <grid-mw-security@nikhef.nl>  
 */

/*!
 * \file glexec_log.h provides the logging functionality.
 */

#ifndef _GLEXEC_LOG_H
#define _GLEXEC_LOG_H

/* Needed for gid_t */
#include <sys/types.h>


/*************************************************************************/
/* LOGGING DEFAULTS							 */
/*************************************************************************/

/* Default syslog facility */
#ifndef GLEXEC_SYSLOG_FACILITY
#define GLEXEC_SYSLOG_FACILITY	LOG_DAEMON
#endif

/* The default log filename when none is specified in glexec.conf */
#ifndef GLEXEC_LOG_FILE
#define GLEXEC_LOG_FILE "/var/log/glexec/glexec_log"
#endif

/* The maximum length that glexec may output for its logging. Longer messages
 * are truncated and will end with ... */
#ifndef GLEXEC_MAX_LOG_LINE
#define GLEXEC_MAX_LOG_LINE 768
#endif

#ifndef GLEXEC_LOG_FILEMASK
#define GLEXEC_LOG_FILEMASK	(S_IRUSR | S_IWUSR | S_IRGRP)
#endif

#ifndef GLEXEC_LOG_DIRMASK
#define GLEXEC_LOG_DIRMASK	(GLEXEC_LOG_FILEMASK | S_IXUSR | S_IXGRP)
#endif

/* Different log levels */
typedef enum {
  GLEXEC_LOG_LEVEL_LOW,	/* Keep this one as the 1st element of this enum */
  GLEXEC_LOG_ERROR,	/* 1 */
  GLEXEC_LOG_WARNING,	/* 2 */
  GLEXEC_LOG_NOTICE,	/* 3 */
  GLEXEC_LOG_INFO,	/* 4 */
  GLEXEC_LOG_DEBUG,	/* 5 */
  GLEXEC_LOG_LEVEL_HIGH   /* Keep this on as the last element of this enum!!! */
} glexec_log_level_t;

/*************************************************************************/
/* FUNCTION PROTOTYPES							 */
/*************************************************************************/

/*!
 * Set logging target to syslog (unsets to file)
 */
void glexec_set_log_to_syslog(void);

/*!
 * Get logging-to-syslog status
 * \return 1 when logging to syslog or 0 otherwise
 */
int glexec_get_log_to_syslog(void);

/*!
 * Set logging target to file (unsets to syslog)
 */
void glexec_set_log_to_file(void);

/*!
 * Get logging-to-file status
 * \return 1 when logging to file or 0 otherwise
 */
int glexec_get_log_to_file(void);

/*!
 * Getter method for glexec logfile gid
 * \return glexec log gid
 */
gid_t glexec_get_loggid(void);

/*!
 * Set the correct syslog facility using specified facility, and reopen when
 * it was already open. When facility is NULL return the current one. Default is
 * GLEXEC_SYSLOG_FACILITY.
 * \param facility *char of facility to use, when NULL don't change.
 * \return facility new integer facility or current when facility == NULL.
 */
int glexec_set_syslog_facility(const char *facility);

/*!
 * Set gLExec loglevel (not LCAS/LCMAPS loglevel)
 * \param loglevel any of  glexec_log_level_t
 *  \param diff_syslog_levels  when 1 glexec_log_levels will be translated into
 *         different syslog levels (only in case logging to syslog of course).
 */
void glexec_set_log_level(glexec_log_level_t loglevel, int diff_syslog_levels);

/*!
 * Glexec log function, automatically opens the log when needed.
 * When logging to file is specified but fails, it will change to syslog.
 * \param loglevel  The log level of this message.
 * \param fmt       The log message itself
 */ 
void glexec_log(glexec_log_level_t loglevel, const char* fmt, ...);

/*!
 * Wrapper function which will try to open the correct log destination
 */
void glexec_open_log(void);

/*!
 * Close syslog and/or logfile
 */
void glexec_close_log(void);

#endif
