/**
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 * glexec_util.c -- helper functions for gLExec
 * \file glexec_util.c
 */



#include "glexec_util.h"

#include "glexec.h"
#include "glexec_log.h"
#include "glexec_account.h"
#include "environ/environ.h"
#include "fileutil/fileutil.h"

#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <ctype.h>
#include <libgen.h>

#include <pwd.h>
#include <grp.h>

/*************************************************************************/
/* START OF GLOBAL VARIABLES                                             */
/*************************************************************************/

static int glexec_switching_mode=-1; /* can be set/get using .... */

/************************************************************************/
/* INTERNAL FUNCTION PROTOTYPES                                         */
/************************************************************************/

/*!
 *  Determine the length of the rsl string including the
 *  terminating '\\0' character.
 *  \param cmd   Executable name
 *  \param argv  All the arguments of the executable.
 *  \return length of the corresponding rsl string.
 */
int glexec_get_rsl_length(const char*, const char**);

/************************************************************************/
/* FUNCTIONS								*/
/************************************************************************/

/*!
 * Sets the internal static glexec_switching_mode variable to the correct glexec
 * mode, either switching or not switching. When we are root, effectively or
 * really, we can switch, otherwise we are in 'logging only' mode. Note that it
 * is only important whether we CAN switch. Note that we also need to check for
 * real uid==0.
 */
void glexec_set_switchingmode(void)  {
    uid_t uid=getuid(),euid=geteuid();

    glexec_switching_mode= ( euid==0 || uid==0 ? 1 : 0);
}

/*!
 * Gets the internal static glexec_switching_mode variable to the correct glexec
 * mode, either switching or not switching. When we are root, effectively or
 * really, we can switch, otherwise we are in 'logging only' mode. Note that it
 * is only important whether we CAN switch. Note that we also need to check for
 * real uid==0.
 * \return 1 when in switching mode, 0 when NOT in switching mode, -1 when undefined
 */
int glexec_is_switching(void)  {
    return glexec_switching_mode;
}

/*!
 * Has the same interface as strdup() but internally it will catch the failures
 * and handle the failures through gmalloc. Hence it only returns NULL when str
 * itself is NULL.
 * \return a duplicate of the str
 */
char * gstrdup(const char *str) {
    size_t siz;
    char *str_out;

    /* Protect against NULL: */
    if (str==NULL)
	return NULL;

    siz = strlen(str) + 1;
    /* Note that gmalloc NEVER returns NULL (it exits), so no need to test for
     * NULL...*/
    str_out = gmalloc(siz);
    
    memcpy(str_out, str, siz);

    return str_out;
}

/*!
 * Has the same interface as strndup(), where available but internally it will
 * catch the failures and handle the failures through gmalloc and is portable.
 * Hence it only returns NULL when str_in itself is NULL.
 * \return a duplicate of not more than n bytes of str_in
 */
char *gstrndup(const char *str_in, size_t n) {
    char *str_out;

    /* Protect against NULL */
    if (str_in==NULL)
	return NULL;

    /* Note that gmalloc NEVER returns NULL (it exits), so no need to test for
     * NULL...*/
    str_out = (char*)gmalloc( (n+1)*sizeof(char) );

    strncpy(str_out, str_in, n);
    str_out[n] = '\0';

    return str_out;
}

/*!
 * Wrapper around malloc() that has the same interface but internally it will
 * catch the failures and handle the failures. Hence it will NEVER return NULL.
 * \return The same as the return value of malloc which is a pointer to
 * allocated data
 */
void *gmalloc(size_t size) {
    void * data = NULL;
    int    curr_errno = 0;

    /* Should protect against malloc-ing zero bytes, it's undefined behaviour */
    if (size==0)    {
	glexec_log(GLEXEC_LOG_ERROR,"Trying to malloc 0 bytes.\n");
        glexec_exit(GLEXEC_ALLOC_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
    }
    data = malloc(size);

    /* Remember the errno before it (might be) overridden by another errno-using
     * (sys)call */
    curr_errno = errno;
    if (!data) {
        glexec_log(GLEXEC_LOG_INFO, "Failed to allocate %d bytes\n", size);
        switch (curr_errno) {
            case ENOMEM :
                glexec_exit(GLEXEC_OUT_OF_MEMORY, GLEXEC_NO_CHILD_EXIT_CODE);
            default :
                glexec_log(GLEXEC_LOG_INFO,
			"malloc() returned with something else then ENOMEM: %d\n",
			curr_errno);
                glexec_exit(GLEXEC_ALLOC_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
        }
    }

    /* Passing the errno state from the malloc function that we're wrapping here
     * */
    errno = curr_errno;
    return data;
}

/*!
 * Wrapper around calloc() that has the same interface but internally it will
 * catch the failures and handle the failures. Hence it will NEVER return NULL.
 * \return The same as the return value of calloc which is a pointer to
 * allocated data
 */
void *gcalloc(size_t nmemb, size_t size) {
    void * data = NULL;
    int    curr_errno = 0;

    /* Should protect against calloc-ing zero bytes, it's undefined behaviour */
    if (nmemb==0 || size==0)    {
	glexec_log(GLEXEC_LOG_ERROR,"Trying to calloc 0 bytes.\n");
        glexec_exit(GLEXEC_ALLOC_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
    }
    data = calloc(nmemb,size);

    /* Remember the errno before it (might be) overridden by another errno-using
     * (sys)call */
    curr_errno = errno;
    if (!data) {
        glexec_log(GLEXEC_LOG_INFO, "Failed to allocate %d bytes\n", nmemb*size);
        switch (curr_errno) {
            case ENOMEM :
                glexec_exit(GLEXEC_OUT_OF_MEMORY, GLEXEC_NO_CHILD_EXIT_CODE);
            default :
                glexec_log(GLEXEC_LOG_INFO,
			"calloc() returned with something else then ENOMEM: %d\n",
			curr_errno);
                glexec_exit(GLEXEC_ALLOC_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
        }
    }

    /* Passing the errno state from the calloc function that we're wrapping here
     * */
    errno = curr_errno;
    return data;
}

/*!
 * Wrapper around realloc() that has the same interface but internally it will
 * catch the failures and handle the failures. Hence it will NEVER return NULL.
 * \return The same as the return value of realloc which is a pointer to
 * allocated data
 */
void *grealloc(void *ptr, size_t size) {
    void * data = NULL;
    int    curr_errno = 0;

    /* Should protect against realloc-ing zero bytes, it's undefined behaviour */
    if (size==0)    {
	glexec_log(GLEXEC_LOG_ERROR,
		   "Trying to realloc 0 bytes: should use free() instead.\n");
        glexec_exit(GLEXEC_ALLOC_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
    }
    data = realloc(ptr,size);

    /* Remember the errno before it (might be) overridden by another errno-using
     * (sys)call */
    curr_errno = errno;
    if (!data) {
        glexec_log(GLEXEC_LOG_INFO, "Failed to allocate %d bytes\n", size);
        switch (curr_errno) {
            case ENOMEM :
                glexec_exit(GLEXEC_OUT_OF_MEMORY, GLEXEC_NO_CHILD_EXIT_CODE);
            default :
                glexec_log(GLEXEC_LOG_INFO,
			"realloc() returned with something else then ENOMEM: %d\n",
			curr_errno);
                glexec_exit(GLEXEC_ALLOC_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
        }
    }

    /* Passing the errno state from the realloc function that we're wrapping here
     * */
    errno = curr_errno;
    return data;
}

/*!
 * Free the memory that is allocated for this list.
 * \param list  List to be freed.
 */
void glexec_free_list(char **list) {
    char **tmp;
    if (list) {
        tmp = list;
        while(*list) {
            free(*list); *list=NULL;
            list++;
        }
        free(tmp); tmp=NULL;
    }
}

/*!
 *  Convert a list of comma separated values into a NULL terminated list
 *  of char pointers. This list will not be generated when either the csv
 *  list is an empty string, no csv list is given (NULL), or when not
 *  enough memory is available to hold the new list.
 *  \param  csv  List of comma separated values.
 *  \return NULL terminated list of char pointers, or NULL in case of
 *          a failure.
 */
char **glexec_convert_csv_to_list(const char *csv) {
  int i, n, l, len;
  int start;
  char **list;

  if (!csv || (csv[0] == '\0')) {
    return NULL;
  }

  /* Count the number of items in the cvs string. */
  for (i=0, n=1; csv[i]; i++) {
    if (csv[i] == ',') {
      n++;
    }
  }

  len = i;   /* Length of the string. */

  /* Allocate enough pointers to hold all items plus one extra for
   * NULL termination of the list. */
  /* note that gcalloc never returns from error, just exits */
  list = (char **)gcalloc(n+1, sizeof(char *));

  /* Start building the list from the string. */
  l = 0;
  for (i=0; i<n; i++) {
    start = l;

    for (; (l<len) && (csv[l] != ','); l++) {}      /* Find the comma.*/

    /* Copy the value over to the list. */
    if (!(list[i] = gstrndup(csv+start, l-start))) {
      /* Could not copy the value, free the list and bail out. */
      glexec_free_list(list);
      return NULL;
    }

    l++;  /* Skip the comma as it is of no use to us anymore.*/

    /* Disregard any space that follow it. */
    for (; (l<len) && isspace(csv[l]); l++) {}
  }

  return list;
}

/*!
 *  Generate an rsl string based on an executable name and its arguments.
 *  \param cmd    Executable name.
 *  \param argv   All the arguments of the executable.
 *  \param rsl    Buffer to store the rsl.
 *  \return Error code.
 */
GLEXEC_ERROR glexec_generate_rsl_string(const char *cmd, const char **argv,
					char **rsl) {
    int length,pos;
    char *buf;

    length = glexec_get_rsl_length(cmd, argv);

    /* gmalloc never returns from error, just exits */
    buf = (char *)gmalloc(sizeof(char) * length);
    
    pos = snprintf(buf, length, "(\\\"executable\\\" = \\\"%s\\\")", cmd);
    if (pos >= length) {
	glexec_log(GLEXEC_LOG_ERROR,
	    "rsl string does not fit into the buffer (%d bytes).\n", length);
	free(buf);
	return GLEXEC_RSL_FAILED;
    }

    while (*argv) {
	if ( (pos += snprintf(buf+pos, length-pos, "(\\\"arguments\\\" = \"%s\")", *argv)) >= length ) {
	    glexec_log(GLEXEC_LOG_ERROR,
		"rsl string does not fit into the buffer (%d bytes).\n", length);
	    free(buf);
	    return GLEXEC_RSL_FAILED;
	}
	argv++;
    }

    *rsl=buf;
    return GLEXEC_SUCCESS;
}

/*!
 *  Determine if a given name matches a pool account
 *
 *  \param pool_name  base name of the pool account
 *  \param user       possible pool account user name
 *  \param n          length of the base name
 *
 *  \return true, if the user matches the pool account, false otherwise.
 *
 */
int glexec_match_pool_account_name(const char *pool_name,const char *user,int n)
{
  /* Match the 'static' part of the pool account name. */
  if ((strlen(user)>n) && strncmp(pool_name, user, n)==0) {
    /* Match the dynamic part, i.e. make sure only numbers are matched. */
    while (user[n] != '\0') {
      if ((user[n] < '0') || (user[n] > '9')) {
        return 0;
      }
      n++;
    }
    return 1;
  }

  return 0;
}

/*!
 *  Read a proxy into a buffer.
 *  \param proxy_file file to read from
 *  \param proxy will contain the proxy
 *  \param lock_mech type of locking mechanism to use
 */
GLEXEC_ERROR glexec_read_proxy(const char* proxy_file,
			       GLEXEC_LOCK_MECH_T lock_mech,
			       char **proxy) {
    int rc;
    int lock_type;

    /* Set correct locking mechanism */
    switch(lock_mech) {
	case GLEXEC_LOCK_FLOCK:	    lock_type=LCK_FLOCK; break;
	case GLEXEC_LOCK_FCNTL:	    lock_type=LCK_FCNTL; break;
	case GLEXEC_LOCK_DISABLED:  lock_type=LCK_NOLOCK; break;
	default:		    return GLEXEC_CONF_FAIL_LOCK_MECH;
    }
    /* Read the proxy */
    rc=cgul_read_proxy(proxy_file, lock_type, proxy);
    /* Parse the return code and log error when needed. */
    switch(rc)	{
	case 0:
	    return GLEXEC_SUCCESS;
	case -1: /* I/O error */
	    glexec_log(GLEXEC_LOG_ERROR,"Failed to read proxyfile %s.\n",
		    proxy_file);
	    glexec_log(GLEXEC_LOG_ERROR,"  errno (%d): %s\n",
		    errno, strerror(errno));
	    return GLEXEC_PROXY_OPEN_FAILED;
	case -2: /* privilege drop error */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Failed to set identity while reading proxy %s.\n",
		    proxy_file);
	    return GLEXEC_COULD_NOT_DROP_PRIVILEGES;
	case -3: /* permission error */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Wrong file permissions for proxy %s.\n",
		    proxy_file);
	    return GLEXEC_WRONG_PERMISSION_PROXY;
	case -4: /* out of memory */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Out of memory while reading proxy %s.\n",
		    proxy_file);
	    return GLEXEC_OUT_OF_MEMORY;
	case -5: /* too many file changes */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "File changed too many times during reading of proxy %s.\n",
		    proxy_file);
	    return GLEXEC_FILE_CHANGED;
	case -6: /* locking error */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Could not lock file during reading of proxy %s.\n",
		    proxy_file);
	    return GLEXEC_LOCK_FAILURE;
	default: /* other or unknown error */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Unknown error while reading proxy %s.\n",
		    proxy_file);
	    return GLEXEC_PROXY_OPEN_FAILED;
    }
}


/************************************************************************/
/* INTERNAL FUNCTIONS							*/
/************************************************************************/

/*!
 *  Determine the length of the rsl string including the
 *  terminating '\\0' character.
 *
 *  \param cmd   Executable name
 *  \param argv  All the arguments of the executable.
 *
 *  \return Length of the corresponding rsl string.
 *
 */
int glexec_get_rsl_length(const char *cmd, const char *argv[])
{
  int i=0, length = strlen(cmd) + 23 + 1;

  while (argv[i]) {
    length += strlen(argv[i]) + 23;
    i++;
  }

  return length;
}
