/**
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 * \file glexec_util.h
 * provides general tools for gLExec
 */

#ifndef _GLEXEC_UTIL_H
#define _GLEXEC_UTIL_H

/* Need to include for env_t */
#include "environ/environ.h"

/* Need for GLEXEC_LOCK_MECH_T */
#include "glexec.h"

/* Need for GLEXEC_ERROR */
#include "glexec_exit.h"
#include <sys/stat.h>

/*************************************************************************/
/* START OF FUNCTION DECLARATIONS                                        */
/*************************************************************************/

/*!
 * Sets whether we are in switching or logging-only mode: switching mode is
 * effective or real uid root.
 */
void glexec_set_switchingmode(void);

/*!
 * Getter function for switching mode.
 * \returns 1 for switching mode, 0 for logging-only mode, -1 for undefined.
 */
int glexec_is_switching(void);

/*!
 * Has the same interface as strdup() but internally it will catch the failures
 * and handle the failures through gmalloc. Hence it only returns NULL when str
 * itself is NULL.
 * \return a duplicate of the str
 */
char *gstrdup(const char *str);

/*!
 * Has the same interface as strndup(), where available but internally it will
 * catch the failures and handle the failures through gmalloc and is portable.
 * Hence it only returns NULL when str_in itself is NULL.
 * \return a duplicate of not more than n bytes of str_in
 */
char *gstrndup(const char *str_in, size_t n);

/*!
 * Wrapper function for malloc: on out-of-memory or size 0 it will exit gLExec,
 * hence it NEVER returns NULL
 * \return size bytes of allocated memory
 */
void *gmalloc(size_t size);

/*!
 * Wrapper function for calloc: on out-of-memory or nmemb*size 0 it will exit
 * gLExec, hence it NEVER returns NULL
 * \return size*nmemb bytes of allocated memory
 */
void *gcalloc(size_t nmemb, size_t size);

/*!
 * Wrapper function for realloc: on out-of-memory or size 0 it will exit gLExec,
 * hence it NEVER returns NULL
 * \return size bytes of allocated memory
 */
void *grealloc(void *ptr,size_t size);

/*!
 *  Free the memory that is allocated for this list.
 * \param list  List to be freed.
 */
void glexec_free_list(char **list);

/*!
 *  Convert a list of comma separated values into a NULL terminated list of char
 *  pointers. This list will not be generated when either the csv list is an
 *  empty string, no csv list is given (NULL), or when not enough memory is
 *  available to hold the new list.
 *  \param  csv  List of comma separated values.
 *  \return NULL terminated list of char pointers, or NULL in case of
 *          a failure.
 */
char **glexec_convert_csv_to_list(const char *csv);

/*!
 * generates a RSL string for given cmd and its arguments
 * \param cmd command
 * \param argv arguments for command
 * \param rsl will contain the RSL
 * \return GLEXEC_SUCCESS when success or a suitable GLEXEC_ERROR
 */
GLEXEC_ERROR glexec_generate_rsl_string(const char *cmd,const char **argv,char **rsl);

/*!
 *  Determine if a given name matches a pool account
 *  \param pool_name  base name of the pool account
 *  \param user       possible pool account user name
 *  \param n          length of the base name
 *  \return true, if the user matches the pool account, false otherwise.
 */
int glexec_match_pool_account_name(const char *, const char *, int);

/*!
 *  Read a proxy into a buffer.
 *  \param proxy_file file to read from
 *  \param proxy will contain the proxy
 *  \param lock_mech type of locking mechanism to use
 */
GLEXEC_ERROR glexec_read_proxy(const char* proxy_file,
				GLEXEC_LOCK_MECH_T lock_mech,
				char **proxy);

#endif
