/**
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 * \file main.c
 */

#ifdef __linux__
#ifdef DEBUG
    #include <sys/prctl.h>
#endif /* DEBUG */
#endif /* __linux__ */

#include "glexec.h"
#include "main_util.h"
#include "glexec_util.h"
#include "glexec_environ.h"
#include "glexec_log.h"
#include "glexec_exit.h"

#include <unistd.h>
#include <stdlib.h>
#include <sys/times.h>
#include <sys/types.h>

/*************************************************************************/
/* START OF main()                                                       */
/*************************************************************************/

int main(int argc, char *argv[]) {
    int glexec_print_defines_mode=0;	/* whether -V is specified */
    char *pem_string= NULL;	/* GLEXEC_CLIENT_CERT */
    char *source_proxy= NULL;	/* GLEXEC_SOURCE_PROXY */
    int t_argc;			/* argc for command to be executed */
    char **t_argv;		/* argv for command to be executed */
    char *target_cwd=NULL;	/* current working directory for target */
    char *fullcmd=NULL;		/* absolute command to be executed */
    char *realcmd=NULL;		/* canonical absolute command to be executed */
    clock_t real_time_start;	/* start time clock tick */
    struct tms tms_start;	/* start time struct */
    pid_t pid;			/* child process pid */
    int rc;			/* exit code for child */
    glexec_account_t caller,ecaller,target; /* real-, effect. caller, target */

#ifdef __linux__
#ifdef DEBUG
    /* echo "1" > /proc/sys/kernel/core_setuid_ok */
    /* echo "1" > /proc/sys/kernel/suid_dumpable */
    prctl(PR_SET_DUMPABLE, 1, 42, 42, 42);
    sleep(5);
#endif /* DEBUG */
#endif /* __linux__ */


#ifdef YES_I_AM_REALLY_SURE_TO_DISABLE_THIS_SECURITY_MEASURE_IN_GLEXEC
    fprintf (stderr, "[gLExec]: Warning: security measures disabled at build time! For testing purposes only, don't use this build in production.\n"); 
#endif /* YES_I_AM_REALLY_SURE_TO_DISABLE_THIS_SECURITY_MEASURE_IN_GLEXEC */

    /* Clean the internal environment, this will also chdir to / */
    glexec_clean_environment();

    /* change to safe directory: user has no more influence... */
    glexec_set_runtime_cwd();

    /* Check number of arguments, simple modes are handled here. Only -V mode
     * has to wait till after glexec_get_caller() is done. */
    glexec_get_runmode(argc, argv, &glexec_print_defines_mode);

    /* initialize switching variable */
    glexec_set_switchingmode();

    /* Read in configuration */
    glexec_read_configuration();

    /* Setup logging */
    glexec_setup_log();

    /* Fill caller structs */
    glexec_get_caller(&caller,&ecaller);

    /* Now we are ready for the 'print_defines' / -V mode if wanted */
    if (glexec_print_defines_mode)
	glexec_print_defines(stdout, &caller);

    /*
     * Check to see if the user running this program is allowed to, based on
     * either whitelist or group membership
     */
    glexec_check_caller_allowed(&caller);

    /* Setup the gLExec internal environment variables */
    glexec_setup_glexec_env(caller.uname);

    /* Read GLEXEC_CLIENT_CERT and GLEXEC_SOURCE_PROXY or its default */
    glexec_get_proxies(&pem_string, &source_proxy);

    /* Set command and argument list for target process */
    t_argc=argc-1;
    t_argv=&argv[1];

    /* Get target account info: will exit upon error.
     * When LCMAPS should do the switch, this is where it's going to happen */
    glexec_auth((const char **)t_argv, pem_string, &target);

    /* Check wether target user may run...*/
    glexec_check_target_allowed(&target);

    /* Stuff to be done when glexec is (configured to be) doing the switch */
    if (glexec_posix_enf_by_glexec()) {
	/* First create a backlog item if needed */
	glexec_create_back_log_item(&caller, &target);

	/* If we are not in logging only mode, do the switch */
	if (glexec_is_switching())
	    glexec_posix_enf(&target);
    } else
	/* Check we are no longer root! */
	glexec_check_no_longer_root(&caller, &target);

    /* Start with a "clean" environment for the target environment. This also
     * tries to chdir to the proper homedir. In any case the new cwd is returnes
     * in target_cwd. Do it AFTER the userwswitch to make sure we have the right
     * permissions. */
    glexec_setup_target_account(&caller,&target,&target_cwd);

    /* Check prog and directory */
    glexec_check_exe_and_dir(t_argv[0],
			     target_cwd, &fullcmd, &realcmd, &caller, &target);

    /* Inform the sysadmin what is being executed */
    glexec_log_argument_list(t_argc, t_argv,
			     target_cwd, fullcmd, realcmd, &caller, &target);

    /* Write proxy (do this only now that only the fork or execve can fail) */
    glexec_write_source_proxy(&caller,&target,&source_proxy);

    /* Fork off a child when needed, otherwise, 0 will be both, also starts
     * timer. */
    pid=glexec_fork(&real_time_start, &tms_start);

    if (pid==0)	{ /* I.e. child */
    	/* Execute with mapped privileges (or non-mapped privs) and from the
	 * child process the following command with specified argv[].
	 * NOTE: We always use the realcmd so that even in case of a symlink
	 * attack we still now which executable has been run. */
        execve(realcmd, t_argv, glexec_get_target_env() );
	
	/* Either NEVER come here or execve() was unsuccessful */
	glexec_execve_exit(target_cwd,fullcmd,realcmd);
    }
    /* Remainder code is the parent! */

    /* Wait for the child to exit */
    rc=glexec_waitpid(real_time_start, &tms_start,pid);

    /* All done: exit */
    glexec_exit(GLEXEC_SUCCESS, rc);

    /* This is to get rid of the warning: "control reaches end of non-void
     * function", which is true, but the compiler doesn't know glexec_exit()
     * never returns. */
    return 0;
}
