/**
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors:
 * 2009-
 *	Oscar Koeroo <okoeroo@nikhef.nl>
 *	Mischa Sall\'e <msalle@nikhef.nl>
 *	NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *	Gerben Venekamp <venekamp@nikhef.nl>
 *	NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 *  \file main_util.h
 *
 *  Provides function definitions for main's helper functions in main_util
 */

#ifndef _MAIN_UTIL_H
#define _MAIN_UTIL_H

/* For glexec_account_t and GLEXEC_LOCK_MECH_T */
#include "glexec.h"
/* For GLEXEC_ERROR */
#include "glexec_exit.h"
/* For env_t */
#include "environ/environ.h"

/* For FILE type */
#include <stdio.h>
/* For uid_t, pid_t */
#include <sys/types.h>

/* For clock_t and struct tms */
#include <sys/times.h>

/************************************************************************/
/* FUNCTION PROTOTYPES                                                  */
/************************************************************************/

/*!
 * Parses the commandline arguments for simple running modes: i.e. no arguments
 * (=error), or a single option argument (-h, -v, -V etc.). In case of -V the
 * parameter print_defines is set. For other simple run modes, the corresponding
 * function (usage, version etc). is run and glexec will exit.
 * \param argc, argv same as for main()
 * \param print_defines set when first option is -V
 */
void glexec_get_runmode(int argc, char *argv[], int *print_defines);

/*!
 * Prints usage text on stdout. exec_name is the executable name.
 */
void glexec_usage(char *exec_name);

/*!
 * Prints build-in defaults on filedescriptor out: -V mode.
 * Only root and special user are allowed, which are compared to the caller.
 * \return GLEXEC_SUCCESS upon success, or GLEXEC_USER_TRIED_V when user is not
 * allowed.
 */
void glexec_print_defines(FILE *out, glexec_account_t *caller_name);

/*!
 * Read the configuration file of glexec.
 *
 * In setuid/switching mode, the configuration file is only read in an
 * unprivileged mode, using the GLEXEC_USER account. If this useraccount does
 * not exist or switching to that account fails an error is returned.
 * Once the configuration file has been read, glexec will resume its previous
 * identity.
 * \return Error code.
 */
GLEXEC_ERROR glexec_read_configuration(void);

/*!
 * Sets up the logging based on the configuration file: destination and
 * loglevel.
 */
void glexec_setup_log(void);

/*!
 * Determines the real and effective caller of gLExec and fills the two account
 * structures with their information. Exits when error is encountered. The
 * fields uid, uname, dir (not for ecaller), gid, gname are set. nsgid and sgid
 * are initialized to -1 and NULL.
 * \param caller real caller
 * \param ecaller effective caller
 */
void glexec_get_caller(glexec_account_t *caller,
		       glexec_account_t *ecaller);

/*!
 * Determines if given caller is allowed to run gLExec based on either explicit
 * whitelisting in the config file, or membership of the special group
 * GLEXEC_WHITELIST_GROUP.
 * \param caller account structure of the gLExec caller
 * \return GLEXEC_SUCCESS when user is allowed or exit when not when not.
 */
GLEXEC_ERROR glexec_check_caller_allowed(glexec_account_t *caller);

/*!
 * sets up the internal gLExec environment, based on the configuration file.
 * \param user calling username
 * \return GLEXEC_SUCCESS on success or a GLEXEC_ERROR.
 */
GLEXEC_ERROR glexec_setup_glexec_env(const char *user);

/*!
 * Reads in GLEXEC_CLIENT_CERT and when needed GLEXEC_SOURCE_PROXY. When
 * GLEXEC_SOURCE_PROXY is not specified, its default value is GLEXEC_CLIENT_CERT
 * which is then just copied in memory.
 * \param pem_string will contain the GLEXEC_CLIENT_CERT
 * \param source_proxy will contain the GLEXEC_SOURCE_PROXY
 * \return GLEXEC_SUCCESS on success or exits upon error
 */
GLEXEC_ERROR glexec_get_proxies(char**pem_string, char **source_proxy);

/*!
 * Does the authentication and authorization for gLExec (LCAS and LCMAPS), based
 * on pem_string (GLEXEC_CLIENT_CERT) and returns a mapped user account in
 * account_info.
 * \param cmd+argv arguments for executable to be run, used to construct a RSL.
 * \param pem_string contains the proxy from GLEXEC_CLIENT_CERT which will be
 *        freed before returning.
 * \param account_info will contain the target account.
 * \return GLEXEC_SUCCESS on success or a GLEXEC_ERROR.
 */
GLEXEC_ERROR glexec_auth(const char **argv, char *pem_string,
	                 glexec_account_t *account_info);

/*!
 * Performs uid/gid checks on the obtained target account:
 * any uid and gid has to be non-zero and >= GLEXEC_UID_MIN/GLEXEC_GID_MIN
 * \param target
 * \return GLEXEC_SUCCESS when successful or exit gLExec.
 */
GLEXEC_ERROR glexec_check_target_allowed(glexec_account_t *target);

/*!
 * Sets up the target account for target. This includes a chdir to the correct
 * homedirectory (for logging-only-mode this will be that of the caller) and
 * setting the env variables PATH, HOME, USER and LOGNAME in addition to those
 * whitelisted in the gLExec configuration file.
 * The environment can be obtained using the glexec_get_target_env() function.
 * \param caller account of the gLExec caller.
 * \param target account of the payload user.
 * \param target_cwd current working directory for payload.
 */
void glexec_setup_target_account(glexec_account_t *caller,
				 glexec_account_t *target,
				 char **target_cwd);

/*!
 * Returns whether gLExec should do the user switch or not.
 */
int glexec_posix_enf_by_glexec(void);

/*!
 * Performs the userswitch.
 * \param target target account
 * Upon error gLExec exits.
 */
void glexec_posix_enf(glexec_account_t *target);

/*!
 *  Create a new backlog item in the backlog_path specified in the config file,
 *  or will return when none is specified. The path used for backlogging needs
 *  to be an absolute path, if not glexec will exit with an error. It will
 *  create a filename based on the absolute path and source user name and then
 *  write the destination user name to it.
 *  \param caller    uid of the source user.
 *  \param target    uid of the destination user.
 *   \return 0 on success, != 0 otherwise.
 */
int glexec_create_back_log_item(glexec_account_t *caller,
                                glexec_account_t *target);

/*!
 * Runs a simple test that gLExec is no longer root.
 * \return effective uid when non-zero or exits when it is.
 */
uid_t glexec_check_no_longer_root(glexec_account_t *caller,
				  glexec_account_t *target);

/*!
 * This function checks the cmd and target_cwd against some security issues: Not
 * set[ug]id, regular file, not changing and optionally the pedantic and
 * no-symlink checks. It obtains the absolute (fullcmd) and canonical absolute
 * (realcmd) path for cmd. In case of problems glexec exists.
 * \param cmd command to be executed
 * \param target_cwd current working directory at payload time
 * \param fullcmd will contain absolute path executable
 * \param realcmd will contain canonical absolute path executable
 * \param caller
 * \param target
 */
void glexec_check_exe_and_dir(char *cmd, char *target_cwd,
			      char **fullcmd, char **realcmd,
			      glexec_account_t *caller,
			      glexec_account_t *target);

/*!
 * Prints out a one-line log entry with all relevant information:
 * real calling username/groupname, target username/groupname, command to be
 * executed and its arguments.
 * \param argc for target process
 * \param argv for target process
 * \param cwd current working directory for target
 * \param fullcmd absolute path for target (can contain symlinks)
 * \param realcmd canonical absolute path for target
 * \param caller real calling user
 * \param target target user
 */
void glexec_log_argument_list(int argc, char *argv[],
			      char *cwd, char *fullcmd, char *realcmd,
			      glexec_account_t *caller,
			      glexec_account_t *target);

/*!
 *  This function writes the proxy for the target user/process, using
 *  target_lock_mechanism.
 *  When GLEXEC_TARGET_PROXY is set, this will be the output filename. When it
 *  is unset, the behaviour depends on the mode: in switching mode, a default
 *  proxy name will be constructed from the uid of the target user and a random
 *  string of characters. In logging only mode, no file will be written.
 *  In all cases X509_USER_PROXY and GLEXEC_TARGET_PROXY are set in the target
 *  environment to the correct proxyfile. In logging only mode without a set
 *  value for GLEXEC_TARGET_PROXY (when no file is written, see above) the value
 *  will be made equal to that of GLEXEC_SOURCE_PROXY or its default
 *  GLEXEC_CLIENT_CERT.
 *  \param caller for caller user
 *  \param target for target user
 *  \param source_proxy contents of GLEXEC_SOURCE_PROXY (or its default), will
 *	   be freed.
 *  \return GLEXEC_SUCCESS on success or a GLEXEC_ERROR.
 */
GLEXEC_ERROR glexec_write_source_proxy(glexec_account_t *caller,
				       glexec_account_t *target,
				       char **source_proxy);

/*!
 * Depending on the linger setting in the configuration file will optionally do
 * a fork() and return the corresponding pid_t. The start time will be set, in
 * linger mode only for the parent. The process for which glexec_fork() returns
 * 0 is supposed to run the payload.
 * \param real_time_start result of times()
 * \param tms_start argument of times()
 * \return non-linger: 0, linger/parent: pid of child, linger/child: 0.
 * Upon error (failing fork) gLExec will exit.
 */
pid_t glexec_fork(clock_t *real_time_start, struct tms *tms_start);

/*!
 * Waits for the child/payload process (in linger mode) to finish and correctly
 * handles, logs etc. the reason (normal exit, signal etc.). Timing statistics
 * are also logged.
 * \param real_time_start start time as set by glexec_fork()
 * \param tms_start start time struct as set by glexec_fork()
 * \param pid pid of child to wait for
 * \return exit code for child process, following (ba)sh convention: when due to
 * a signal: 128+<signal>, in case of unsignalled but abnormal program end, we
 * return 1. See http://tldp.org/LDP/abs/html/exitcodes.html for details.
 */ 
int glexec_waitpid(clock_t real_time_start, struct tms *tms_start, pid_t pid);

#endif
